# -*- coding: utf-8 -*-
"""
Created on Thu Feb 15 21:53:45 2024

@author: dvezinet
"""
# Built-in
import os


# Standard
# import matplotlib.pyplot as plt


# spectrally-specific
from ._setup_teardown import setup_module0, teardown_module0
from .._class02_SpectralFit import SpectralFit as Collection
# from .._saveload import load
from . import _spectralfit_input as _inputs
from . import _hxr_pulses


_PATH_HERE = os.path.dirname(__file__)
_PATH_INPUT = os.path.join(_PATH_HERE, 'input')
_PATH_OUTPUT = os.path.join(_PATH_HERE, 'output')


#######################################################
#
#     Setup and Teardown
#
#######################################################


def setup_module(module):
    setup_module0(module)


def teardown_module(module):
    teardown_module0(module)


#######################################################
#
#     Instanciate and populate
#
#######################################################


class Test00_Populate():

    # ------------------------
    #   setup and teardown
    # ------------------------

    @classmethod
    def setup_class(cls):
        pass

    def setup_method(self):

        # instanciate
        self.coll = Collection()

        # add data
        _inputs.add_data(self.coll)

        # add spectral lines
        pfe_json = os.path.join(_PATH_INPUT, 'spectrallines.json')
        self.coll.add_spectral_lines_from_file(pfe_json)

    # ------------------------
    #   Populating
    # ------------------------

    # ---------------
    # 1d spectral fit

    def test01_add_spectral_fit_1d_single(self):
        # add spectral fit 1d
        _inputs.add_fit_single(self.coll)

    def test02_add_spectral_fit_1d_double(self):
        # add spectral fit 1d
        _inputs.add_fit_double(self.coll)

    def test03_add_spectral_fit_1d_multi(self):
        # add spectral fit 1d
        _inputs.add_fit_multi(self.coll)

    def test04_plot_spectral_fit_input_validity_1d(self):
        # plot 1d
        _inputs.plot_input_validity(self.coll, key_data='data1d')

    def test05_compute_spectral_fit_1d_single(self):
        # compute 1d
        _inputs.compute_fit_single(self.coll)

    def test06_compute_spectral_fit_1d_double(self):
        # compute 1d
        _inputs.compute_fit_double(self.coll)

    def test07_compute_spectral_fit_1d_single_binning(self):
        # compute 1d
        _inputs.compute_fit_single(self.coll, binning=10)

    def test08_compute_spectral_fit_1d_multi(self):
        # compute 1d
        _inputs.compute_fit_multi(self.coll, key_data='data1d')

    # ---------------
    # 2d spectral fit

    # def test06_add_spectral_fit_2d(self):
    #     # add spectral fit 2d
    #     _inputs.add_fit(self.coll, key_data='data2d')

    # def test07_plot_spectral_fit_input_validity_2d(self):
    #     # plot 2d
    #     _inputs.plot_input_validity(self.coll, key_data='data2d')

    # ------------------------
    #   Special cases
    # ------------------------

    def test99_compute_HXR_pulses(self):
        _hxr_pulses.main(self.coll)


# -*- coding: utf-8 -*-
"""
Created on Wed Jul 24 09:36:24 2024

@author: dvezinet
"""


# common
import numpy as np


#############################################
#############################################
#       DEFAULTS
#############################################


_LFIT_ORDER = [
    'sol',
    'model', 'data', 'sigma', 'lamb', 'bs',
    'positive', 'nsigma', 'fraction',
    'mask', 'domain', 'focus',
]


#############################################
#############################################
#       Show
#############################################


def _show(coll=None, which=None, lcol=None, lar=None, show=None):

    # ---------------------------
    # column names
    # ---------------------------

    lcol.append([which] + _LFIT_ORDER)

    # ---------------------------
    # data
    # ---------------------------

    lkey = [
        k1 for k1 in coll._dobj.get(which, {}).keys()
        if show is None or k1 in show
    ]

    lar0 = []
    for k0 in lkey:

        # initialize with key
        arr = [k0]

        # add nb of func of each type
        dfit = coll.dobj[which][k0]
        for k1 in _LFIT_ORDER:

            if k1 in ['model', 'data', 'sigma', 'lamb', 'bs', 'sol']:
                nn = '' if dfit[f"key_{k1}"] is None else str(dfit[f"key_{k1}"])

            elif k1 in ['nsigma', 'fraction', 'positive']:
                nn = str(dfit['dvalid'][k1])

            elif k1 in ['mask']:
                nn = str(dfit['dvalid']['mask']['key'] is not None)

            elif k1 in ['domain']:
                c0 = all([
                    len(v0['spec']) == 1
                    and np.allclose(v0['spec'][0], np.inf*np.r_[-1, 1])
                    for k0, v0 in dfit['dvalid']['domain'].items()
                ])
                if c0:
                    nn = ''
                else:
                    lk = list(dfit['dvalid']['domain'].keys())
                    if len(lk) == 2 and lk[0] != dfit['key_lamb']:
                        lk = [lk[1], lk[0]]

                    nn = ', '.join([
                        str(len(dfit['dvalid']['domain'][k0]['spec']))
                        for k0 in lk
                    ])

            elif k1 in ['focus']:
                if dfit['dvalid'].get('focus') is None:
                    nn = ''
                else:
                    nn = len(dfit['dvalid']['focus'])
                    nn = f"{nn} / {dfit['dvalid']['focus_logic']}"

            arr.append(nn)

        lar0.append(arr)

    lar.append(lar0)

    return lcol, lar


# #!/usr/bin/python3
# -*- coding: utf-8 -*-


# standard
import os
import itertools as itt
import json


# common
import numpy as np
import scipy.constants as scpct


_OPENADAS_ONLINE = True


_SPECTRAL_DUNITS = {
    'wavelength': ['m', 'mm', 'um', 'nm', 'pm', 'A'],
    'energy': ['TeV', 'GeV', 'MeV', 'keV', 'eV', 'J'],
    'frequency': ['THz', 'GHz', 'MHz', 'kHz', 'Hz'],
}


# ###############################################################
# ###############################################################
#                       from openadas
# ###############################################################


def from_openadas(
    lambmin=None,
    lambmax=None,
    element=None,
    charge=None,
    online=None,
    update=None,
    create_custom=None,
    dsource0=None,
    dref0=None,
    ddata0=None,
    dobj0=None,
    which_lines=None,
):
    """
    Load lines and pec from openadas, either:
        - online = True:  directly from the website
        - online = False: from pre-downloaded files in ~/.spectrally/openadas/

    Provide wavelengths in m

    Example:
    --------
            >>> import spectrally as sp
            >>> coll = sp.Collection()
            >>> coll.add_spectral_lines_from_openadas(
                element='Mo',
                lambmin=3.94e-10,
                lambmax=4e-10,
            )

    """

    # Preliminary import and checks
    from .openadas import _requests
    from .openadas import _read_files

    if online is None:
        online = _OPENADAS_ONLINE

    # Load from online if relevant
    if online is True:
        try:
            out = _requests.step01_search_online_by_wavelengthA(
                lambmin=lambmin*1e10,
                lambmax=lambmax*1e10,
                element=element,
                charge=charge,
                verb=False,
                returnas=np.ndarray,
                resolveby='file',
            )
            lf = sorted(set([oo[0] for oo in out]))
            out = _requests.step02_download_all(
                files=lf,
                update=update,
                create_custom=create_custom,
                verb=False,
            )
        except Exception as err:
            msg = (
                str(err) + "\n\n"
                "For some reason data could not be downloaded from openadas\n"
                "\t=> see error message above\n"
                "\t=> maybe check your internet connection?\n"
            )
            raise Exception(msg)

    # Load for local files
    dne, dte, dpec, lion, dsource, dlines = _read_files.step03_read_all(
        lambmin=lambmin,
        lambmax=lambmax,
        element=element,
        charge=charge,
        pec_as_func=False,
        format_for_DataStock=True,
        dsource0=dsource0,
        dref0=dref0,
        ddata0=ddata0,
        dlines0=None if dobj0 is None else dobj0.get(which_lines),
        verb=False,
    )

    # ---------------
    # dref - Te + ne

    dref = dte
    dref.update(dne)

    # ------------
    # ddata - pec

    ddata = dpec

    # ------------------
    # lambda0 => lamb0

    lk = list(dlines.keys())
    for k0 in lk:
        dlines[k0]['lamb0'] = dlines[k0]['lambda0']
        del dlines[k0]['lambda0']

    # ------------------
    # ions and source

    # Only keep ions and sources not already stored
    if dobj0 is not None and dobj0.get('ion') is not None:
        lion = [k0 for k0 in lion if k0 not in dobj0['ion'].keys()]

    if dobj0 is not None and dobj0.get('source') is not None:
        dsource = {
            k0: v0 for k0, v0 in dsource.items()
            if k0 not in dobj0['source'].keys()
        }

    # dobj (lines, ion, source)
    dobj = {
        which_lines: dlines,
        'ion': {k0: {} for k0 in lion},
        'source': dsource,
    }

    return ddata, dref, dobj


# ###############################################################
# ###############################################################
#                       from nist
# ###############################################################


def from_nist(
    lambmin=None,
    lambmax=None,
    element=None,
    charge=None,
    ion=None,
    wav_observed=None,
    wav_calculated=None,
    transitions_allowed=None,
    transitions_forbidden=None,
    cache_from=None,
    cache_info=None,
    verb=None,
    create_custom=None,
    dsource0=None,
    dlines0=None,
    group_lines=None,
):
    """
    Load lines from nist, either:
        - cache_from = False:  directly from the website
        - cache_from = True: from pre-downloaded files in ~/.spectrally/nist/

    Provide wavelengths in m

    Example:
    --------
            >>> import spectrally as sp
            >>> coll = sp.Collection()
            >>> sp.add_spectral_lines_from_nist(
                element='Mo',
                lambmin=3.94e-10,
                lambmax=4e-10,
            )

    """

    # Preliminary import and checks
    from .nist import _requests

    if verb is None:
        verb = False
    if cache_info is None:
        cache_info = False

    # Load from online if relevant
    dlines, dsources = _requests.step01_search_online_by_wavelengthA(
        element=element,
        charge=charge,
        ion=ion,
        lambmin=lambmin*1e10,
        lambmax=lambmax*1e10,
        wav_observed=wav_observed,
        wav_calculated=wav_calculated,
        transitions_allowed=transitions_allowed,
        transitions_forbidden=transitions_forbidden,
        info_ref=True,
        info_conf=True,
        info_term=True,
        info_J=True,
        info_g=True,
        cache_from=cache_from,
        cache_info=cache_info,
        return_dout=True,
        return_dsources=True,
        verb=verb,
        create_custom=create_custom,
        format_for_DataStock=True,
        dsource0=dsource0,
        dlines0=dlines0,
    )

    # ------------------
    # lambda0 => lamb0

    lk = list(dlines.keys())
    for k0 in lk:
        dlines[k0]['lamb0'] = dlines[k0]['lambda0']
        del dlines[k0]['lambda0']

    # --------
    # ions
    lion = sorted(set([dlines[k0]['ion'] for k0 in dlines.keys()]))

    # ------------
    # dobj (lines)

    dobj = {
        group_lines: dlines,
        'ion': {k0: {} for k0 in lion},
        'source': dsources,
    }
    return dobj


# ###############################################################
# ###############################################################
#                       from file
# ###############################################################


def from_file(coll=None, pfe=None):

    # ---------------
    # check file
    # --------------

    pfe = _check_pfe(pfe)

    # ---------------
    # open file
    # ---------------

    if pfe.endswith('.json'):
        with open(pfe) as json_file:
            dlines = json.load(json_file)

    elif pfe.endswith('.npz'):
        dlines = dict(np.load(pfe, allow_pickle=True))

    elif pfe.endswith('.py'):
        dlines = _from_module(pfe=pfe)

    # ions
    lions = sorted(set([
        v0.get('ion') for v0 in dlines.values()
        if v0.get('ion') is not None
    ]))

    # sources
    lsources = sorted(set([
        v0.get('source') for v0 in dlines.values()
        if v0.get('source') is not None
    ]))

    dobj = {
        coll._which_lines: dlines,
        'ion': {k0: {} for k0 in lions},
        'source': {k0: {} for k0 in lsources},
    }

    return dobj


#############################################
#       check and extract
#############################################


def _check_pfe(pfe):

    # ------------------
    # basic check
    # ------------------

    if not (isinstance(pfe, str) and os.path.isfile(pfe)):
        msg = (
            "Arg pfe must be a str, a valid 'path/file.ext'\n"
            f"Provided:\n\t{pfe}\n"
        )
        raise Exception(msg)

    pfe = os.path.abspath(pfe)

    # ------------
    # extension
    # ------------------

    if pfe.endswith('.npz'):
        pass

    elif pfe.endswith('.json'):
        pass

    elif pfe.endswith('.py'):
        pass

    else:
        msg = (
            "Unhandled pfe extension!\n"
            "Usable extensions: '.npz', '.json', '.py'\n"
            f"Provided:\n\t{pfe}\n"
        )
        raise Exception(msg)

    return pfe


def _check_extract_dict_from_mod(mod, k0):
    lk1 = [
        k0, k0.upper(),
        '_'+k0, '_'+k0.upper(),
        '_d'+k0, '_D'+k0.upper(),
        'd'+k0, 'D'+k0.upper(),
        k0+'s', k0.upper()+'S'
        '_d'+k0+'s', '_D'+k0.upper()+'S',
        'd'+k0+'s', 'D'+k0.upper()+'S',
    ]
    lk1 = [k1 for k1 in lk1 if hasattr(mod, k1)]
    if len(lk1) > 1:
        msg = "Ambiguous attributes: {}".format(lk1)
        raise Exception(msg)
    elif len(lk1) == 0:
        return

    if hasattr(mod, lk1[0]):
        return getattr(mod, lk1[0])
    else:
        return


def _from_module(pfe=None):

    # ------------
    # Check input

    c0 = (
        os.path.isfile(pfe)
        and pfe[-3:] == '.py'
    )
    if not c0:
        msg = (
            "\nProvided Path-File-Extension (pfe) not valid!\n"
            + "\t- expected: absolute path to python module\n"
            + "\t- provided: {}".format(pfe)
        )
        raise Exception(msg)
    pfe = os.path.abspath(pfe)

    # -----------
    # Load module

    path, fid = os.path.split(pfe)
    import importlib.util
    spec = importlib.util.spec_from_file_location(fid[:-3], pfe)
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)

    # -----------------------------------------
    # extract source, transition, ion, element

    dobj = {}
    for k0 in ['source', 'transition', 'ion', 'element']:
        dd = _check_extract_dict_from_mod(mod, k0)
        if dd is not None:
            dobj[k0] = dd

    # --------
    # add ion

    if 'ion' not in dobj.keys():
        lions = np.array([
                v0['ion'] for k0, v0 in mod.dlines.items()
                if 'ion' in v0.keys()
        ]).ravel()
        if len(lions) > 0:
            dobj['ion'] = {
                k0: {'ion': k0} for k0 in lions
            }
        else:
            lIONS = np.array([
                    v0['ION'] for k0, v0 in mod.dlines.items()
                    if 'ION' in v0.keys()
            ]).ravel()
            if len(lIONS) > 0:
                dobj['ION'] = {
                    k0: {'ION': k0} for k0 in lIONS
                }

    return mod.dlines


# ##############################################################
# ##############################################################
#                       Units conversion
# ##############################################################


def _check_convert_spectral(
    data_in=None,
    units_in=None, units_out=None,
    returnas=None,
):

    # returnas
    if returnas is None:
        returnas = 'data'
    if returnas not in ['data', 'coef']:
        msg = (
            """
            Arg return as must be:
            - 'data': return the converted data
            - 'coef': return the conversion coefficient
            """
        )
        raise Exception(msg)

    # data_in
    if data_in is None:
        if returnas == 'data':
            msg = "If returnas='data', arg data cannot be None!"
            raise Exception(msg)

    else:

        if not isinstance(data_in, np.ndarray):
            try:
                data_in = np.asarray(data_in)
            except Exception as err:
                msg = "Arg data shall be convertible to a np.ndarray!"
                raise Exception(msg)

        if data_in.dtype not in [int, float]:
            msg = (
                "Arg data must be a np.ndarray of dtype int or float!\n"
                f"\t- data.dtype = {data_in.dtype.name}\n"
            )
            raise Exception(msg)

    # units
    units = list(
        itt.chain.from_iterable([vv for vv in _SPECTRAL_DUNITS.values()])
    )
    if units_in not in units or units_out not in units:
        msg = (
            """
            Both units_in and units_out must be in:
            - {}
            - {}
            - {}

            Provided:
            - units_in: {}
            - units_out: {}
            """.format(
                'wavelength: {}'.format(_SPECTRAL_DUNITS['wavelength']),
                'energy: {}'.format(_SPECTRAL_DUNITS['energy']),
                'frequency: {}'.format(_SPECTRAL_DUNITS['frequency']),
                units_in, units_out,
            )
        )
        raise Exception(msg)

    return data_in, returnas


def _convert_spectral_coef(units_in=None, units_out=None):
    """ Get conversion coef """
    k0_in = [k0 for k0, v0 in _SPECTRAL_DUNITS.items() if units_in in v0][0]
    k0_out = [k0 for k0, v0 in _SPECTRAL_DUNITS.items() if units_out in v0][0]

    if units_in == units_out:
        return 1., False

    # ---------
    # First case: same category

    inv = False
    if k0_in == k0_out:
        indin = _SPECTRAL_DUNITS[k0_in].index(units_in)
        indout = _SPECTRAL_DUNITS[k0_out].index(units_out)

        if k0_in == 'frequency':
            coef = 10**(3*(indout-indin))

        elif k0_in == 'wavelength':
            if units_in == 'A':
                coef = 10**(3*(indout-(indin-1)) + 2)
            elif units_out == 'A':
                coef = 10**(3*((indout-1)-indin) - 2)
            else:
                coef = 10**(3*(indout-indin))

        elif k0_in == 'energy':
            if units_in == 'J':
                coef = 10**(3*(indout-(indin-1))) / scpct.e
            elif units_out == 'J':
                coef = 10**(3*((indout-1)-indin)) * scpct.e
            else:
                coef = 10**(3*(indout-indin))

    # ---------
    # For each category, convert to reference (m, eV, Hz)
    else:

        # coefs_in
        if k0_in == 'wavelength':
            # units_in -> eV
            coef_in, _ = _convert_spectral_coef(
                units_in=units_in, units_out='m',
            )
        elif k0_in == 'energy':
            coef_in, _ = _convert_spectral_coef(
                units_in=units_in, units_out='J',
            )
        elif k0_in == 'frequency':
            coef_in, _ = _convert_spectral_coef(
                units_in=units_in, units_out='Hz',
            )

        # coefs_out
        if k0_out == 'wavelength':
            # units_in -> eV
            coef_out, _ = _convert_spectral_coef(
                units_in='m', units_out=units_out,
            )
        elif k0_out == 'energy':
            coef_out, _ = _convert_spectral_coef(
                units_in='J', units_out=units_out,
            )
        elif k0_out == 'frequency':
            coef_out, _ = _convert_spectral_coef(
                units_in='Hz', units_out=units_out,
            )

        # ------------------
        # Cross combinations between (m, J, Hz)

        # E = h*f = h*c/lambda
        if k0_in == 'wavelength':
            inv = True
            if k0_out == 'energy':
                # m -> J
                coef_cross = scpct.h * scpct.c
            elif k0_out == 'frequency':
                # m -> Hz
                coef_cross = scpct.c

        elif k0_in == 'energy':
            if k0_out == 'wavelength':
                # J -> m
                inv = True
                coef_cross = scpct.h * scpct.c
            elif k0_out == 'frequency':
                # J -> Hz
                coef_cross = 1./scpct.h

        elif k0_in == 'frequency':
            if k0_out == 'wavelength':
                # Hz -> m
                inv = True
                coef_cross = scpct.c
            elif k0_out == 'energy':
                # Hz -> J
                coef_cross = scpct.h

        if inv:
            coef = coef_cross*coef_out / coef_in
        else:
            coef = coef_in*coef_cross*coef_out

    return coef, inv


def convert_spectral(
    data_in=None,
    units_in=None, units_out=None,
    returnas=None,
):
    """ convert wavelength / energy/ frequency

    Available units:
        wavelength: m, mm, nm, A
        energy:     J, eV, keV
        frequency:  Hz, kHz, MHz, GHz
    """

    # Check inputs
    data_in, returnas = _check_convert_spectral(
        data_in=data_in,
        units_in=units_in, units_out=units_out,
        returnas=returnas,
    )

    # Convert
    # k0_in = [k0 for k0, v0 in _SPECTRAL_DUNITS.items() if units_in in v0][0]
    # k0_ou = [k0 for k0, v0 in _SPECTRAL_DUNITS.items() if units_ou in v0][0]

    # trivial case first
    if units_in == units_out:
        return data_in

    coef, inv = _convert_spectral_coef(units_in=units_in, units_out=units_out)
    if returnas == 'data':
        if inv:
            return coef / data_in
        else:
            return coef * data_in
    else:
        return coef, inv


#!/usr/bin/env python

# Built-in
import sys
import os
import warnings


_HERE = os.path.abspath(os.path.dirname(__file__))
_PKG = 'spectrally'


# import parser dict
sys.path.insert(1, _HERE)
from _dparser import _DPARSER
_ = sys.path.pop(1)


###################################################
###################################################
#       Main
###################################################


def main():
    # Parse input arguments

    # Parse arguments
    ddef, parser = _DPARSER['version']()

    # Parse arguments
    args = parser.parse_args()

    # Call function
    get_version(ddef=ddef, **dict(args._get_kwargs()))

    return


###################################################
###################################################
#       function
###################################################


def get_version(
    verb=None,
    envvar=None,
    path=None,
    warn=None,
    force=None,
    ddef=None,
):
    """ Print tofu version and / or store in environment variable """

    # --------------
    # Check inputs
    # --------------

    # extract input
    kwd = locals()
    for k0 in set(ddef.keys()).intersection(kwd.keys()):
        if kwd[k0] is None:
            kwd[k0] = ddef[k0]
    verb, envvar, path = kwd['verb'], kwd['envvar'], kwd['path']
    warn, force = kwd['warn'], kwd['force']

    # --------------
    # Check
    # --------------

    # verb, warn, force
    dbool = {'verb': verb, 'warn': warn, 'force': force}
    for k0, v0 in dbool.items():
        if v0 is None:
            dbool[k0] = ddef[k0]
        if not isinstance(dbool[k0], bool):
            msg = (
                f"Arg {k0} must be a bool\n"
                "\t- provided: {dbool[k0]}\n"
            )
            raise Exception(msg)

    # --------------
    # Check inputs
    # --------------

    # set envvar
    if isinstance(envvar, bool):
        if envvar is True:
            envvar = ddef['name']

    elif isinstance(envvar, str):
        envvar = envvar.upper()

    else:
        msg = (
            "Arg envvar must be either:\n"
            f"\t- None:  set to default ({ddef['envvar']})\n"
            "\t- False: no setting of environment variable\n"
            f"\t- True:  set default env. var. ({ddef['name']})\n"
            "\t- str:   set provided env. variable\n\n"
            f" you provided: {envvar}"
        )
        raise Exception(msg)

    # warning on name of env variable
    if envvar is not False and envvar in os.environ.keys():
        if dbool['warn'] is True:
            msg = (
                "Chosen environment variable name already exists:\n"
                f"\t- {envvar}: {os.environ[envvar]}\n"
                "  => Are you sure you want to change it?\n"
                f"       use force=True to overwrite {envvar}\n"
                "       use warn=False to disable this warning"
            )
            warnings.warn(msg)

    # --------------
    # Fetch version from git tags, and write to version.py
    # Also, when git is not available (PyPi package), use stored version.py
    # --------------

    pfe = os.path.join(path, 'version.py')
    if not os.path.isfile(pfe):
        msg = (
            f"It seems your current {_PKG} install has no version.py:\n"
            f"\t- looked for: {pfe}"
        )
        raise Exception(msg)

    # --------------
    # Read file
    # --------------

    with open(pfe, 'r') as fh:
        version = fh.read().strip().split("=")[-1].replace("'", '')
    version = version.lower().replace('v', '').replace(' ', '')

    # --------------
    # Outputs
    # --------------

    if dbool['verb'] is True:
        print(version)
    if envvar is not False:
        c0 = ((envvar in os.environ.keys() and dbool['force'] is True)
              or envvar not in os.environ.keys())
        if c0 is True:
            os.environ[envvar] = version

    return


###################################################
###################################################
#       __main__
###################################################


if __name__ == '__main__':
    main()



"""
This module contains tests for tofu.geom in its structured version
"""


# Built-in
import os
import shutil


__all__ = ['setup_module0', 'teardown_module0']


#######################################################
#
#     DEFAULTS
#
#######################################################


_PATH_HERE = os.path.dirname(__file__)
_PATH_OUTPUT = os.path.join(_PATH_HERE, 'output')


_PKG = 'spectrally'
_PATH_SP = os.path.join(os.path.expanduser('~'), f'.{_PKG}')


_CUSTOM = os.path.dirname(_PATH_HERE)
_CUSTOM = os.path.join(_CUSTOM, 'scripts', f'{_PKG}custom.py')


#######################################################
#
#     Setup and Teardown
#
#######################################################


def clean_output(path=_PATH_OUTPUT):
    """ Remove all temporary output files that may have been forgotten """
    lf = [
        ff for ff in os.listdir(path)
        if ff.endswith('.npz')
        or ff.endswith('.json')
    ]
    if len(lf) > 0:
        for ff in lf:
            os.remove(os.path.join(path, ff))


def create_local_path():

    # clean
    if os.path.isdir(_PATH_SP):
        shutil.rmtree(_PATH_SP)

    # re-create
    os.system(f'python {_CUSTOM}')


def setup_module0(module):
    create_local_path()


def teardown_module0(module):
    clean_output()
    # clean
    if os.path.isdir(_PATH_SP):
        shutil.rmtree(_PATH_SP)



# -*- coding: utf-8 -*-
"""
Created on Wed Jul 24 09:36:24 2024

@author: dvezinet
"""


# built-in
import itertools as itt


# common
import numpy as np
import datastock as ds


# local
from . import _class01_model_dict as _model_dict


__all__ = ['get_available_spectral_model_functions']


#############################################
#############################################
#       Show
#############################################


def _show(coll=None, which=None, lcol=None, lar=None, show=None):

    # ---------------------------
    # list of functions
    # ---------------------------

    # list of models
    lkey = [
        k1 for k1 in coll._dobj.get(which, {}).keys()
        if show is None or k1 in show
    ]

    # list of relevant functions
    lfunc = []
    for k0 in lkey:
        dmod = coll.dobj[which][k0]['dmodel']
        for k1 in _model_dict._LMODEL_ORDER:
            lk2 = [k2 for k2, v2 in dmod.items() if v2['type'] == k1]
            if len(lk2) > 0 and k1 not in lfunc:
                lfunc.append(k1)

    # reorder
    lfunc = [k1 for k1 in _model_dict._LMODEL_ORDER if k1 in lfunc]

    # ---------------------------
    # column names
    # ---------------------------

    lcol.append([which] + lfunc + ['constraints', 'free var'])

    # ---------------------------
    # data array
    # ---------------------------

    lar0 = []
    for k0 in lkey:

        # initialize with key
        arr = [k0]

        # add nb of func of each type
        dmod = coll.dobj[which][k0]['dmodel']
        for k1 in lfunc:
            nn = str(len([k2 for k2, v2 in dmod.items() if v2['type'] == k1]))
            arr.append(nn)

        # add nb of constraints
        dconst = coll.dobj[which][k0]['dconstraints']['dconst']
        nn = str(len([k1 for k1, v1 in dconst.items() if len(v1) > 1]))
        arr.append(nn)

        # add number of free variables
        lfree = coll.get_spectral_model_variables(k0, returnas='free')['free']
        lall = coll.get_spectral_model_variables(k0, returnas='all')['all']
        arr.append(f"{len(lfree)} / {len(lall)}")

        lar0.append(arr)

    lar.append(lar0)

    return lcol, lar


#############################################
#############################################
#       Show single model
#############################################


def _show_details(coll=None, key=None, lcol=None, lar=None, show=None):

    # ---------------------------
    # get dmodel
    # ---------------------------

    wsm = coll._which_model
    dmodel = coll.dobj[wsm][key]['dmodel']
    dconst = coll.dobj[wsm][key]['dconstraints']['dconst']

    lkeys = coll.dobj[wsm][key]['keys']
    llvar = [dmodel[kf]['var'] for kf in lkeys]

    nvarmax = np.max([len(lvar) for lvar in llvar])
    lfree = coll.get_spectral_model_variables(key, returnas='free')['free']

    lpar = sorted(set(itt.chain.from_iterable([
        v0.get('param', {}).keys() for v0 in dmodel.values()
    ])))

    # ---------------------------
    # column names
    # ---------------------------

    lvar = [f"var{ii}" for ii in range(nvarmax)]
    lcol.append(['func', 'type', ' '] + lvar + [' '] + lpar)

    # ---------------------------
    # data
    # ---------------------------

    lar0 = []
    for kf in lkeys:

        # initialize with key, type
        arr = [kf, dmodel[kf]['type'], '|']

        # add variables of each func
        for ii, k1 in enumerate(dmodel[kf]['var']):
            key = f"{kf}_{k1}"
            if key in lfree:
                nn = key
            else:
                gg = [kg for kg, vg in dconst.items() if key in vg.keys()][0]
                nn = f"{key}({dconst[gg]['ref']})"

            arr.append(nn)

        # complement
        arr += ['' for ii in range(nvarmax - ii - 1)] + ['|']

        # add parameters of each func
        for k1 in lpar:
            nn = dmodel[kf].get('param', {}).get(k1, '')
            if not isinstance(nn, str):
                nn = f"{nn:.6e}"
            arr.append(nn)

        lar0.append(arr)

    lar.append(lar0)

    return lcol, lar


#############################################
#############################################
#       Show available models
#############################################


def get_available_spectral_model_functions(
    # print parameters
    sep=None,
    line=None,
    justify=None,
    table_sep=None,
    # bool options
    verb=None,
    returnas=None,
):
    """ Print all te available functions types for spectral fitting

    All arguments are optional
    >>> get_available_spectral_model_functions()


    Parameters
    ----------
    sep : str, optional
        seperator character
    line : str, optional
        line character
    justify : str, optional
        'left' or 'right'
    table_sep : str, optional
        separator character between tables
    verb : bool, optional
        Whether to print
    returnas : bool, optional
        Whether to return

    Returns
    -------

    May return the printed msg as str if requested

    """

    # ---------------------------
    # get dmodel
    # ---------------------------

    dmodel = _model_dict._DMODEL

    lkeys = _model_dict._LMODEL_ORDER
    llvar = [dmodel[kf]['var'] for kf in lkeys]

    nvarmax = np.max([len(lvar) for lvar in llvar])
    lpar = sorted(set(itt.chain.from_iterable([
        [kk[0] for kk in v0.get('param', [])] for v0 in dmodel.values()
    ])))

    # ---------------------------
    # column names
    # ---------------------------

    lvar = [f"var{ii}" for ii in range(nvarmax)]
    lcol = [
        ['use case', ' ', 'func type', ' ', 'description', ' ']
        + lvar + [' '] + lpar
    ]

    # ---------------------------
    # data
    # ---------------------------

    lar = []
    lar0 = []
    for kf in lkeys:

        # use case
        if kf == 'poly':
            use = 'bck'
        elif kf == 'gauss':
            use = 'lines'
        elif kf == 'pulse_exp':
            use = 'pulses'
        else:
            use = ''

        # initialize with key, type
        arr = [use, '|', kf, '|', dmodel[kf]['description'], '|']

        # add variables of each func
        for ii, k1 in enumerate(dmodel[kf]['var']):
            arr.append(k1)

        # complement
        arr += ['' for ii in range(nvarmax - ii - 1)] + ['|']

        # add parameters of each func
        for k1 in lpar:
            ln = [
                vv[1] for vv in dmodel[kf].get('param', [])
                if vv[0] == k1
            ]
            if len(ln) > 0:
                if hasattr(ln[0], '__name__'):
                    nn = ln[0].__name__
                else:
                    nn = f"{ln[0]}"
            else:
                nn = ""
            arr.append(nn)

        lar0.append(arr)

    lar.append(lar0)

    # ---------------------------
    # pretty print
    # ---------------------------

    return ds._generic_utils.pretty_print(
        headers=lcol,
        content=lar,
        sep=sep,
        line=line,
        justify=justify,
        table_sep=table_sep,
        verb=verb,
        returnas=returnas,
    )


'''

Tutorial script for spectrally

Shows both a SXR (gaussian spectrum) and HXR (pulse shape) example

cjperks
July 16, 2024

'''

# built-in
import os


# common
import numpy as np
import scipy.constants as scpct


__all__ = ['main']


# ######################################################
# ######################################################
#               DEFAULTS
# ######################################################


_PATH_HERE = os.path.dirname(__file__)
_PATH_INPUT = os.path.join(os.path.dirname(_PATH_HERE), 'tests', 'input')


# ######################################################
# ######################################################
#               Main
# ######################################################


def main(
    data='ebit',
    compute=True,
    chain=None,
    solver=None,
    binning=None,
):
    """ Tutorial on how to use spectrally

    Two examples 'sxr' and 'hxr' are available
    Both are fitted using several different models and constraints


    Parameters
    ----------
    data : str, optional
        Flag indicating which example to run. The default is 'sxr'.
        - 'sxr': a double-gaussian spectrum with poly background
        - 'hxr': a series of experimental pulse shapes

    Returns
    -------
    coll : Collection instance
        The Collection instance containing all inputs, fit models and data

    """

    import spectrally as sp

    # ------------------------
    # instanciate a Collection
    # ------------------------

    coll = sp.Collection()

    # make sure data is lower-case
    data = data.lower()

    # -----------------
    # create wavelength vector and time-dependent gaussian spectrum
    # -----------------

    _add_data(coll, data=data)

    # -----------------
    # print available spectral model functions
    # -----------------

    sp.get_available_spectral_model_functions()

    # -----------------
    # display a figure showing details of each spectral model function
    # -----------------

    if data in ['sxr', 'ebit']:
        _ = sp.display_spectral_model_function('gauss')

    elif data == 'hxr':
        _ = sp.display_spectral_model_function('pulse_exp')
        _ = sp.display_spectral_model_function('pulse_gauss')
        _ = sp.display_spectral_model_function('lognorm')

    # ------------------
    # add spectral model
    # ------------------

    _add_spectral_model(coll=coll, data=data)

    # ------------------
    # add spectral fit (check input data validity, but not compute yet)
    # ------------------

    _add_spectral_fit(coll=coll, data=data)

    # ---------------------
    # plot validity
    # ---------------------

    if data == 'sxr':
        dax = coll.plot_spectral_fit_input_validity('sf1')
    elif data == 'ebit':
        dax = coll.plot_spectral_fit_input_validity('sf0')
    else:
        dax = coll.plot_spectral_fit_input_validity('sf_exp')

    # ---------------------
    # compute spectral fit
    # ---------------------

    dax = None
    dout = None
    if compute is True:

        _compute_spectral_fit(
            coll=coll,
            data=data,
            chain=chain,
            solver=solver,
            binning=binning,
        )

        # ---------------------
        # plot fit
        # ---------------------

        if data == 'sxr':
            dax = coll.plot_spectral_fit('sf1')
        elif data == 'ebit':
            dax = coll.plot_spectral_fit('sf0')
        else:
            dax = coll.plot_spectral_fit('sf_lognorm')

        # ---------------------
        # extract moments
        # ---------------------

        dout = None
        if data == 'sxr':
            dout = coll.get_spectral_model_moments('sf1')
        elif data == 'ebit':
            dout = coll.get_spectral_model_moments('sf0')
        else:
            dout = coll.get_spectral_model_moments('sf_exp')

    return coll, dax, dout


# ######################################################
# ######################################################
#               add data
# ######################################################


def _add_data(coll=None, data=None):

    # ------------
    # SXR
    # ------------

    if data == 'sxr':

        # ------------
        # lamb

        nlamb = 300
        lamb = np.linspace(3.94e-10, 4.e-10, nlamb)

        # ------------
        # time vector

        nt = 50
        t = np.linspace(0, 10, nt)

        # -----------
        # spectra

        lamb0 = 3.96e-10 + 0.001e-10 * np.cos(t[:, None])
        lamb1 = 3.98e-10 + 0.001e-10 * np.sin(t[:, None])


        # poly
        poly = (1 + np.cos(t)[:, None]) * 15 * np.ones((1, nlamb))

        gauss0 = (
            (1 + np.cos(t)[:, None]) * 50
            * np.exp(-(lamb[None, :]-lamb0)**2/(2*0.001e-10**2))
        )

        gauss1 = (
            (1 + np.sin(t)[:, None]) * 80
            * np.exp(-(lamb[None, :]-lamb1)**2/(2*0.0005e-10**2))
        )


        spectra = (
            poly
            + gauss0
            + gauss1
            + np.random.random((nt, nlamb)) * 5
        )

        # --------------------
        # populate collection

        # refs
        coll.add_ref('nlamb', size=nlamb)
        coll.add_ref('nt', size=nt)

        # data - lamb
        coll.add_data(
            'lamb',
            data=lamb,
            ref='nlamb',
            units='m',
        )


        # data - t
        coll.add_data(
            't',
            data=t,
            ref='nt',
            units='s',
        )


        # data - spectrum
        coll.add_data(
            'spectra',
            data=spectra,
            ref=('nt', 'nlamb'),
            units='counts',
        )

    # ------------
    # HXR
    # ------------

    elif data == 'hxr':

        # --------------
        # load file

        # pfe = path/file.ext
        pfe = os.path.join(_PATH_INPUT, 'HXR_pulses.npz')

        # load
        dout = dict(np.load(pfe, allow_pickle=True))

        # -------------
        # ref

        coll.add_ref('npulse', size=dout['pulse'].size)
        coll.add_ref('nsamp', size=dout['data'].shape[0])

        # -------------
        # data

        # pulses
        coll.add_data(
            key='pulses',
            data=dout['pulse'],
            ref='npulse',
            units='index',
        )

        # indices
        coll.add_data(
            key='sample',
            data=np.arange(0, dout['data'].shape[0]),
            ref='nsamp',
            units='index',
        )

        # data
        coll.add_data(
            key='current',
            data=dout['data'],
            ref=('nsamp', 'npulse'),
            units='A',
        )

    # ------------
    # EBIT
    # ------------

    elif data == 'ebit':

        # --------------
        # load file

        # pfe = path/file.ext
        pfe = os.path.join(_PATH_INPUT, 'EBIT_spectrally_test.npz')

        # load
        dout = dict(np.load(pfe, allow_pickle=True))['arr_0'].tolist()

        # -------------
        # ref

        coll.add_ref('nv', size=dout['Vs'].size)
        coll.add_ref('nlamb', size=dout['lambs'].size)

        # -------------
        # data

        # voltages
        coll.add_data(
            key='voltages',
            data=dout['Vs'],
            ref='nv',
            units='kV',
        )

        # wavelength
        coll.add_data(
            key='lamb',
            data=dout['lambs'],
            ref='nlamb',
            units='m',
        )

        # data
        coll.add_data(
            key='spectra',
            data=dout['data'],
            ref=('nv', 'nlamb'),
            units='counts',
        )

        # --------------
        # add spectral lines

        coll.add_spectral_line(
            key='l0',
            ion='Ar16+',
            lamb0=2.719e-10,
            source='user',
        )

        coll.add_spectral_line(
            key='l1',
            ion='Ar15+',
            lamb0=2.726e-10,
            source='user',
        )

        coll.add_spectral_line(
            key='l2',
            ion='Ar16+',
            lamb0=2.735e-10,
            source='user',
        )

    # ------------
    # Error
    # ------------

    else:

        msg = "Unknown data for tutorial"
        raise Exception(msg)

    return


# ######################################################
# ######################################################
#               spectral model
# ######################################################


def _add_spectral_model(coll=None, data=None):

    # ------------
    # SXR
    # ------------

    if data == 'sxr':

        # argon mass
        mz = scpct.m_u * 39.9

        # first mode : bck + 2 gauss
        coll.add_spectral_model(
            key='sm0',
            dmodel={
                'bck': 'poly',
                'l0': {'type': 'gauss', 'lamb0': 3.96e-10, 'mz': mz},
                'l1': {'type': 'gauss', 'lamb0': 3.98e-10, 'mz': mz},
            },
        )

        # first mode : bck with 0 slope + 2 gauss with same width
        coll.add_spectral_model(
            key='sm1',
            dmodel={
                'bck': 'poly',
                'l0': {'type': 'gauss', 'lamb0': 3.96e-10, 'mz': mz},
                'l1': {'type': 'gauss', 'lamb0': 3.98e-10, 'mz': mz},
            },
            dconstraints={
                # 'g0': {
                #     'ref': 'l0_sigma',
                #     'l1_sigma': [0, 1, 0],
                #     },
                'g2': {
                    'ref': 'bck_a0',
                    'bck_a1': [0, 0, 0],
                    'bck_a2': [0, 0, 0],
                },
            },
        )

    # ------------
    # HXR
    # ------------

    elif data == 'hxr':

        # constraints for all models: the poly back has a slope of 0
        dconstraints = {
            'g0': {
                'ref': 'bck_a0',
                'bck_a1': [0, 0, 0],
                'bck_a2': [0, 0, 0],
            },
        }

        # first model : bck with only offset + pulse_exp
        coll.add_spectral_model(
            key='sm_exp',
            dmodel={
                'bck': 'poly',
                'p0': {'type': 'pulse_exp'},
            },
            dconstraints=dconstraints,
        )

        # first model : bck with only offset + pulse_gauss
        coll.add_spectral_model(
            key='sm_gauss',
            dmodel={
                'bck': 'poly',
                'p0': {'type': 'pulse_gauss'},
            },
            dconstraints=dconstraints,
        )

        # first model : bck with only offset + lognorm
        coll.add_spectral_model(
            key='sm_lognorm',
            dmodel={
                'bck': 'poly',
                'p0': {'type': 'lognorm'},
            },
            dconstraints=dconstraints,
        )

    elif data == 'ebit':

        # argon mass
        mz = scpct.m_u * 39.9

        # first mode : bck with 0 slope + 2 gauss with same width
        coll.add_spectral_model(
            key='sm0',
            dmodel={
                'bck': 'poly',
                'l0': {'type': 'gauss'},
                'l1': {'type': 'gauss'},
                'l2': {'type': 'gauss'},
            },
            dconstraints={
                'g0': {
                    'ref': 'bck_a0',
                    'bck_a1': [0, 0, 0],
                    'bck_a2': [0, 0, 0],
                },
                'g1': {
                    'ref': 'l0_sigma',
                    'l1_sigma': [0, 1, 0],
                    'l2_sigma': [0, 1, 0],
                },
            },
        )

    return


# ######################################################
# ######################################################
#               spectral fits
# ######################################################


def _add_spectral_fit(coll=None, data=None):

    # ------------
    # SXR
    # ------------

    if data == 'sxr':

        # first spectral model
        coll.add_spectral_fit(
            key='sf0',
            key_model='sm0',
            key_data='spectra',
            key_lamb='lamb',
            key_sigma='poisson',
            dvalid={
                'nsigma': 3,
                'fraction': 0.31,
            },
        )

        # second spectral model
        coll.add_spectral_fit(
            key='sf1',
            key_model='sm1',
            key_data='spectra',
            key_lamb='lamb',
            key_sigma='5%',
            dvalid={
                'nsigma': 3,
                'fraction': 0.31,
            },
        )

        # second spectral model again (for later use)
        coll.add_spectral_fit(
            key='sf2',
            key_model='sm1',
            key_data='spectra',
            key_lamb='lamb',
            key_sigma=5,
            dvalid={
                'nsigma': 3,
                'fraction': 0.31,
            },
        )

    # ------------
    # HXR
    # ------------

    elif data == 'hxr':

        for k0 in coll.dobj['spect_model'].keys():
            coll.add_spectral_fit(
                key=k0.replace('sm_', 'sf_'),
                key_model=k0,
                key_data='current',
                key_sigma='poisson',
                key_lamb='sample',
                # params
                dvalid={
                    'nsigma': 2,
                    'fraction': 0.11,
                },
            )

    # ------------
    # EBIT
    # ------------

    elif data == 'ebit':

        for k0 in coll.dobj['spect_model'].keys():
            coll.add_spectral_fit(
                key='sf0',
                key_model=k0,
                key_data='spectra',
                key_sigma='poisson',
                key_lamb='lamb',
                # params
                dvalid={
                    'nsigma': 2,
                    'fraction': 0.11,
                },
            )

            coll.add_spectral_fit(
                key='sf1',
                key_model=k0,
                key_data='spectra',
                key_sigma='poisson',
                key_lamb='lamb',
                absolute_sigma=False,
                # params
                dvalid={
                    'nsigma': 2,
                    'fraction': 0.11,
                },
            )

            coll.add_spectral_fit(
                key='sf2',
                key_model=k0,
                key_data='spectra',
                key_sigma='10%',
                key_lamb='lamb',
                absolute_sigma=False,
                # params
                dvalid={
                    'nsigma': 2,
                    'fraction': 0.11,
                },
            )

            coll.add_spectral_fit(
                key='sf3',
                key_model=k0,
                key_data='spectra',
                key_sigma=10,
                key_lamb='lamb',
                absolute_sigma=False,
                # params
                dvalid={
                    'nsigma': 2,
                    'fraction': 0.11,
                },
            )

    return


# ######################################################
# ######################################################
#               compute
# ######################################################


def _compute_spectral_fit(
    coll=None,
    data=None,
    chain=None,
    solver=None,
    binning=None,
):

    # ------------
    # check inputs
    # ------------

    if chain is None:
        if data == 'sxr':
            chain = True
        else:
            chain = False

    # ------------
    # SXR
    # ------------

    if data == 'sxr':

        for k0 in ['sf0', 'sf1']:
            coll.compute_spectral_fit(
                k0,
                solver=solver,
                dsolver_options={'nfev': 1000},
                chain=chain,
                verb=2,
                strict=True,
            )
        # ---------------------
        # compute spectral fit with user-provided scales, bounds and x0
        # ---------------------

        # Focus on the second spectral model again
        # this time the bounds and initial of free variablesis set by the user
        # (otherwise they are set automatically

        # we first extract the list of free variables
        xfree = coll.get_spectral_model_variables('sm1', returnas='free')['free']
        print(xfree)
        # ['bck_a0', 'l0_amp', 'l0_vccos', 'l0_sigma', 'l1_amp', 'l1_vccos']

        # Build a dict for lower bounds
        dbounds_low = {
            'bck_a0': 0,
            'l0_amp': 0,
            'l0_vccos': -0.03 / 3.96,
            'l0_sigma': 0.0001e-10,
            'l1_amp': 0,
            'l1_vccos': -0.03 / 3.98,
        }

        # Build a dict for upper bounds
        dbounds_up = {
            'bck_a0': 50,  # 25
            'l0_amp': 200,
            'l0_vccos': 0.03 / 3.96,
            'l0_sigma': 0.005e-10,
            'l1_amp': 200,
            'l1_vccos': 0.03 / 3.98,
        }

        # Build a dict for inital guesses
        dx0 = {
            'bck_a0': 20,
            'l0_amp': 100,
            'l0_vccos': 0. / 3.96, # 0.02 / 3.96
            'l0_sigma': 0.001e-10,
            'l1_amp': 100,
            'l1_vccos': 0. / 3.98, # -0.02 / 3.98
        }

        # second spectral model with user-provided scales, bounds and x0
        coll.compute_spectral_fit(
            'sf2',
            # user-provided
            dbounds_low=dbounds_low,
            dbounds_up=dbounds_up,
            dx0=dx0,
            # options
            solver=solver,
            chain=chain,
            verb=2,
            strict=True,
        )

    # ------------
    # HXR
    # ------------

    else:

        for k0 in coll.dobj['spect_fit'].keys():
            coll.compute_spectral_fit(
                k0,
                solver=solver,
                dsolver_options={'nfev': 1000},
                chain=chain,
                binning=binning,
                verb=2,
                strict=True,
            )

    return


# ######################################################
# ######################################################
#               __main__
# ######################################################


if __name__ == "__main__":
    coll = main()


# -*- coding: utf-8 -*-
"""
Created on Sat Jul  6 17:31:49 2024

@author: dvezinet
"""


import numpy as np


from . import _class01_model_dict


#############################################
#############################################
#    DEFAULTS
#############################################


_DEF_SCALES_FACTORS = {
    'width': 2,   # lambD / 2
    'shift': 2,   # lambD / 2
}


_DEF_X0_FACTORS = {
    'width': 10,  # lambD / 10
    'shift': 10,  # lambD / 10
}


_DEF_X0 = {
    'lognorm': {
        'sigma': 1.,
    },
}


#############################################
#############################################
#    harmonize dict for scale, bounds, x0
#############################################


def _get_dict(
    lx_free_keys=None,
    dmodel=None,
    din=None,
    din_name=None,
):

    # --------------
    # trivial
    # --------------

    if din is None:
        return {}

    # --------------
    # non-trivial
    # --------------

    c0 = (
        isinstance(din, dict)
        and all([isinstance(k0, str) for k0 in din.keys()])
    )
    if not c0:
        msg = (
            f"Arg '{din_name}' must be a dict of the form:\n"
            "\t- key_of_free_variable: value (float)\n"
            "Provided:\n{din}"
        )
        raise Exception(msg)

    # --------------
    # check keys
    # --------------

    derr = {
        k0: v0 for k0, v0 in din.items()
        if k0 not in lx_free_keys or not np.isscalar(v0)
    }
    if len(derr) > 0:
        lstr = [f"\t- '{k0}': {v0}" for k0, v0 in derr.items()]
        msg = (
            "The following key / values are non-conform from '{din_name}':/n"
            + "\n".join(lstr)
            + "\nAll keys must be natching free variable names!\n"
            "Available keys:\n{lk_free_keys}"
        )
        raise Exception(msg)

    # --------------
    # convert to ind / val format
    # --------------

    dout = {}
    for k0, v0 in din.items():

        # get func name and type + variable name
        ftype = dmodel['_'.join(k0.split('_')[:-1])]['type']
        if not ftype in _class01_model_dict._DMODEL.keys():
            msg = (
                "Unknown function type for variable:\n"
                f"\t- var: {k0}\n"
                f"\t- ftype: {ftype}\n"
            )
            raise Exception(msg)


        var = k0.split('_')[-1]

        if ftype not in dout.keys():
            dout[ftype] = {}

        if var not in dout[ftype].keys():
            dout[ftype][var] = {'ind': [], 'val': []}

        dout[ftype][var]['ind'].append(lx_free_keys.index(k0))
        dout[ftype][var]['val'].append(v0)

    # --------------
    # sort
    # --------------

    lktypes = list(dout.keys())
    for k0 in lktypes:
        lkvar = list(dout[k0].keys())
        for k1 in lkvar:
            inds = np.argsort(dout[k0][k1]['ind'])
            dout[k0][k1]['ind'] = np.array(dout[k0][k1]['ind'])[inds]
            dout[k0][k1]['val'] = np.array(dout[k0][k1]['val'])[inds]

    return dout


#############################################
#############################################
#       get scales, bounds
#############################################


def _get_scales_bounds(
    nxfree=None,
    lamb=None,
    data=None,
    iok=None,
    axis=None,
    dind=None,
    dscales=None,
    dbounds_low=None,
    dbounds_up=None,
    # binning
    dbinning=None,
):

    # ------------------
    # initialize
    # ------------------

    scales = np.zeros((nxfree,), dtype=float)
    bounds0 = np.zeros((nxfree,), dtype=float)
    bounds1 = np.zeros((nxfree,), dtype=float)

    # ------------------
    # prepare
    # ------------------

    (
        lamb0, lambD, lambm,
        lamb_amax, lamb_amin,
        lamb_ext,
        data_max, data_min,
        data_mean, data_median,
        data_pulse_sign,
    ) = _prepare(
        lamb=lamb,
        data=data,
        iok=iok,
        axis=axis,
        # binning
        dbinning=dbinning,
    )

    ldins = [(dscales, scales), (dbounds_low, bounds0), (dbounds_up, bounds1)]

    # ------------------
    # all poly
    # ------------------

    kfunc = 'poly'
    if dind.get(kfunc) is not None:

        # -------
        # a0

        kvar = 'a0'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            scales[ival] = max(np.abs(data_max), np.abs(data_min))
            bounds0[ival] = -10.
            bounds1[ival] = 10.

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=None if ii == 0 else scales
                )

        # -------
        # a1

        kvar = 'a1'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            scales[ival] = max(np.abs(data_max), np.abs(data_min))
            bounds0[ival] = -10.
            bounds1[ival] = 10.

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=None if ii == 0 else scales
                )

        # -------
        # a2

        kvar = 'a2'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            scales[ival] = max(np.abs(data_max), np.abs(data_min))
            bounds0[ival] = -10.
            bounds1[ival] = 10.

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=None if ii == 0 else scales
                )

    # --------------------
    # all exponentials
    # --------------------

    kfunc = 'exp_lamb'
    if dind.get(kfunc) is not None:

        rate = np.nanmax([
            np.abs(
                np.log(data_max * lamb[0] / (data_min * lamb[-1]))
                / (1/lamb[-1] - 1./lamb[0])
            ),
            np.abs(
                np.log(data_min * lamb[0] / (data_max * lamb[-1]))
                / (1/lamb[-1] - 1./lamb[0])
            ),
        ])

        # rate
        kvar = 'rate'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            scales[ival] = rate
            bounds0[ival] = 0.
            bounds1[ival] = 10.

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=None if ii == 0 else scales
                )

        # amp
        kvar = 'amp'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            scales[ival] = data_mean * np.exp(rate/lambm) * lambm
            bounds0[ival] = 0.
            bounds1[ival] = 10.

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=None if ii == 0 else scales
                )

    # -----------------
    # all gaussians
    # -----------------

    kfunc = 'gauss'
    if dind.get(kfunc) is not None:

        # amp
        kvar = 'amp'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            scales[ival] = data_max - data_min
            bounds0[ival] = 0.
            bounds1[ival] = 10.

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=None if ii == 0 else scales
                )

        # vccos
        kvar = 'vccos'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            scales[ival] = (lambD/lambm) / _DEF_SCALES_FACTORS['shift']
            bounds0[ival] = -1.
            bounds1[ival] = 1.

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=None if ii == 0 else scales
                )

        # sigma
        kvar = 'sigma'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            scales[ival] = lambD / _DEF_SCALES_FACTORS['width']
            bounds0[ival] = 1e-3
            bounds1[ival] = 2

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=None if ii == 0 else scales
                )

    # -----------------
    # all lorentz
    # -----------------

    kfunc = 'lorentz'
    if dind.get(kfunc) is not None:

        # amp
        kvar = 'amp'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            scales[ival] = data_max - data_min
            bounds0[ival] = 0.
            bounds1[ival] = 10.

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=None if ii == 0 else scales
                )

        # vccos
        kvar = 'vccos'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            scales[ival] = (lambD/lambm) / _DEF_SCALES_FACTORS['shift']
            bounds0[ival] = -1.
            bounds1[ival] = 1.

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=None if ii == 0 else scales
                )

        # gam
        kvar = 'gam'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            scales[ival] = lambD / _DEF_SCALES_FACTORS['width']
            bounds0[ival] = 1e-3
            bounds1[ival] = 2

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=None if ii == 0 else scales
                )

    # -----------------
    # all pvoigt
    # -----------------

    kfunc = 'pvoigt'
    if dind.get(kfunc) is not None:

        # amp
        kvar = 'amp'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            scales[ival] = data_max - data_min
            bounds0[ival] = 0.
            bounds1[ival] = 10.

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=None if ii == 0 else scales
                )

        # vccos
        kvar = 'vccos'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            scales[ival] = (lambD/lambm) / _DEF_SCALES_FACTORS['shift']
            bounds0[ival] = -1.
            bounds1[ival] = 1.

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=None if ii == 0 else scales
                )

        # sigma
        kvar = 'sigma'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            scales[ival] = lambD / _DEF_SCALES_FACTORS['width']
            bounds0[ival] = 1e-3
            bounds1[ival] = 2

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=None if ii == 0 else scales
                )

        # gam
        kvar = 'gam'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            scales[ival] = lambD / _DEF_SCALES_FACTORS['width']
            bounds0[ival] = 1e-3
            bounds1[ival] = 2

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=None if ii == 0 else scales
                )

    # -----------------
    # all voigt
    # -----------------

    kfunc = 'voigt'
    if dind.get(kfunc) is not None:

        # amp
        kvar = 'amp'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            scales[ival] = data_max - data_min
            bounds0[ival] = 0.
            bounds1[ival] = 10.

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=None if ii == 0 else scales
                )

        # vccos
        kvar = 'vccos'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            scales[ival] = (lambD/lambm) / _DEF_SCALES_FACTORS['shift']
            bounds0[ival] = -1.
            bounds1[ival] = 1.

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=None if ii == 0 else scales
                )

        # sigma
        kvar = 'sigma'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            scales[ival] = lambD / _DEF_SCALES_FACTORS['width']
            bounds0[ival] = 1e-3
            bounds1[ival] = 2

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=None if ii == 0 else scales
                )

        # gam
        kvar = 'gam'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            scales[ival] = lambD / _DEF_SCALES_FACTORS['width']
            bounds0[ival] = 1e-3
            bounds1[ival] = 2

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=None if ii == 0 else scales
                )

    # ------------------
    # all pulse_exp
    # ------------------

    kfunc = 'pulse_exp'
    if dind.get(kfunc) is not None:

        _get_scales_bounds_pulse(
            dind=dind,
            kfunc=kfunc,
            ldins=ldins,
            # data
            data_min=data_min,
            data_max=data_max,
            data_pulse_sign=data_pulse_sign,
            # lamb
            lamb=lamb,
            lambm=lambm,
            lambD=lambD,
            # arrays to fill
            scales=scales,
            bounds0=bounds0,
            bounds1=bounds1,
        )

    # ------------------
    # all pulse_gauss
    # ------------------

    kfunc = 'pulse_gauss'
    if dind.get(kfunc) is not None:

        _get_scales_bounds_pulse(
            dind=dind,
            kfunc=kfunc,
            ldins=ldins,
            # data
            data_min=data_min,
            data_max=data_max,
            data_pulse_sign=data_pulse_sign,
            # lamb
            lamb=lamb,
            lambm=lambm,
            lambD=lambD,
            # arrays to fill
            scales=scales,
            bounds0=bounds0,
            bounds1=bounds1,
        )

    # ------------------
    # all lognorm
    # ------------------

    kfunc = 'lognorm'
    if dind.get(kfunc) is not None:

        # useful for guessing

        # max at t - t0 = exp(mu - sigma**2)
        # max = amp * exp(sigma**2/2 - mu)
        # variance = (exp(sigma**2) - 1) * exp(2mu + sigma**2)
        # => mu = 0.5 * (log(std**2 / (exp(sigma**2) - 1)) - sigma**2)
        # skewness = (exp(sigma**2) + 2) * sqrt(exp(sigma**2) - 1)

        sigma = _DEF_X0[kfunc]['sigma']

        std = lambD / 5
        std_min = lambD / 100
        std_max = lambD

        mu = 0.5 * (np.log(std**2/(np.exp(sigma**2) - 1)) - sigma**2)
        mu_abs = np.abs(mu)
        mu_min = 0.5 * (np.log(std_min**2/(np.exp(sigma**2) - 1)) - sigma**2)
        mu_max = 0.5 * (np.log(std_max**2/(np.exp(sigma**2) - 1)) - sigma**2)

        # sigma
        kvar = 'sigma'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            scales[ival] = sigma
            bounds0[ival] = 0.1
            bounds1[ival] = 5

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=None if ii == 0 else scales
                )

        # mu
        kvar = 'mu'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            scales[ival] = mu_abs
            bounds0[ival] = mu_min / mu_abs
            bounds1[ival] = mu_max / mu_abs

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=None if ii == 0 else scales
                )

        # amp
        # max = amp * exp(sigma**2/2 - mu)
        kvar = 'amp'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            scales[ival] = (data_max - data_min) * np.exp(mu - 0.5*sigma**2)
            if data_pulse_sign > 0:
                bounds0[ival] = 0.
                bounds1[ival] = 10.
            else:
                bounds0[ival] = -10.
                bounds1[ival] = 0.

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=None if ii == 0 else scales
                )

        # tau
        # max at lamb - (lamb00 + lambD * tau) = exp(mu - sigma**2)
        kvar = 'tau'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            scales[ival] = 1
            bounds0[ival] = 0
            bounds1[ival] = 1

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=None if ii == 0 else scales
                )

    return scales, bounds0, bounds1


#############################################
#############################################
#       get x0
#############################################


def _get_x0(
    nxfree=None,
    lamb=None,
    data=None,
    iok=None,
    axis=None,
    dind=None,
    dx0=None,
    scales=None,
    # binning
    dbinning=None,
):

    # ------------------
    # initialize
    # ------------------

    x0 = np.zeros((nxfree,), dtype=float)

    # ------------------
    # prepare
    # ------------------

    (
        lamb0, lambD, lambm,
        lamb_amax, lamb_amin,
        lamb_ext,
        data_max, data_min,
        data_mean, data_median,
        data_pulse_sign,
    ) = _prepare(
        lamb=lamb,
        data=data,
        iok=iok,
        axis=axis,
        # binning
        dbinning=dbinning,
    )

    ldins = [(dx0, x0)]

    # ------------------
    # all poly
    # ------------------

    kfunc = 'poly'
    if dind.get(kfunc) is not None:

        # a1max = ((data_max - data_min) / lambD) / 10

        # -------
        # a0

        kvar = 'a0'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            x0[ival] = data_mean / scales[ival]

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=scales,
                )

        # -------
        # a1

        kvar = 'a1'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            x0[ival] = 0. / scales[ival]

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=scales,
                )

        # -------
        # a2

        kvar = 'a2'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            x0[ival] = 0. / scales[ival]

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=scales,
                )

    # --------------------
    # all exponentials
    # ------------------

    kfunc = 'exp_lamb'
    if dind.get(kfunc) is not None:

        rate = np.nanmax([
            np.abs(
                np.log(data_max * lamb[0] / (data_min * lamb[-1]))
                / (1/lamb[-1] - 1./lamb[0])
            ),
            np.abs(
                np.log(data_min * lamb[0] / (data_max * lamb[-1]))
                / (1/lamb[-1] - 1./lamb[0])
            ),
        ]) / 10.

        # rate
        kvar = 'rate'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            x0[ival] = rate / scales[ival]

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=scales,
                )

        # amp
        kvar = 'amp'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            scales[ival] = data_mean * np.exp(rate/lambm) * lambm / scales[ival]

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=scales,
                )

    # -----------------
    # all gaussians
    # -----------------

    kfunc = 'gauss'
    if dind.get(kfunc) is not None:

        # amp
        kvar = 'amp'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            x0[ival] = (data_max - data_min) / 2 / scales[ival]

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=scales,
                )

        # vccos
        kvar = 'vccos'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            x0[ival] = lambD / _DEF_X0_FACTORS['shift'] / scales[ival]

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=scales,
                )

        # sigma
        kvar = 'sigma'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            x0[ival] = lambD / _DEF_X0_FACTORS['width'] / scales[ival]

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=scales,
                )

    # -----------------
    # all lorentz
    # -----------------

    kfunc = 'lorentz'
    if dind.get(kfunc) is not None:

        # amp
        kvar = 'amp'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            x0[ival] = (data_max - data_min) / 2 / scales[ival]

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=scales,
                )

        # vccos
        kvar = 'vccos'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            x0[ival] = lambD / _DEF_X0_FACTORS['shift'] / scales[ival]

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=scales,
                )

        # gam
        kvar = 'gam'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            x0[ival] = lambD / _DEF_X0_FACTORS['width'] / scales[ival]

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=scales,
                )

    # -----------------
    # all pvoigt
    # -----------------

    kfunc = 'pvoigt'
    if dind.get(kfunc) is not None:

        # amp
        kvar = 'amp'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            x0[ival] = (data_max - data_min) / 2. / scales[ival]

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=scales,
                )

        # vccos
        kvar = 'vccos'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            x0[ival] = lambD / _DEF_X0_FACTORS['shift'] / scales[ival]

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=scales,
                )

        # sigma
        kvar = 'sigma'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            x0[ival] = lambD / _DEF_X0_FACTORS['width'] / scales[ival]

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=scales,
                )

        # gam
        kvar = 'gam'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            x0[ival] = lambD / _DEF_X0_FACTORS['width'] / scales[ival]

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=scales,
                )

    # -----------------
    # all voigt
    # -----------------

    kfunc = 'voigt'
    if dind.get(kfunc) is not None:

        # amp
        kvar = 'amp'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            x0[ival] = (data_max - data_min) / 2 / scales[ival]

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=scales,
                )

        # vccos
        kvar = 'vccos'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            x0[ival] = lambD / _DEF_X0_FACTORS['shift'] / scales[ival]

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=scales,
                )

        # sigma
        kvar = 'sigma'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            x0[ival] = lambD / _DEF_X0_FACTORS['width'] / scales[ival]

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=scales,
                )

        # gam
        kvar = 'gam'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            x0[ival] = lambD / _DEF_X0_FACTORS['width'] / scales[ival]

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=scales,
                )

    # ------------------
    # all pulse_exp
    # ------------------

    kfunc = 'pulse_exp'
    if dind.get(kfunc) is not None:

        _get_x0_pulse(
            dind=dind,
            kfunc=kfunc,
            ldins=ldins,
            # data
            data_min=data_min,
            data_max=data_max,
            data_pulse_sign=data_pulse_sign,
            # lamb
            lamb0=lamb0,
            lambm=lambm,
            lambD=lambD,
            lamb_ext=lamb_ext,
            # arrays to fill
            x0=x0,
            scales=scales,
        )

    # ------------------
    # all pulse_gauss
    # ------------------

    kfunc = 'pulse_gauss'
    if dind.get(kfunc) is not None:

        _get_x0_pulse(
            dind=dind,
            kfunc=kfunc,
            ldins=ldins,
            # data
            data_min=data_min,
            data_max=data_max,
            data_pulse_sign=data_pulse_sign,
            # lamb
            lamb0=lamb0,
            lambm=lambm,
            lambD=lambD,
            lamb_ext=lamb_ext,
            # arrays to fill
            x0=x0,
            scales=scales,
        )

    # ------------------
    # all lognorm
    # ------------------

    kfunc = 'lognorm'
    if dind.get(kfunc) is not None:

        # useful for guessing
        sigma = _DEF_X0[kfunc]['sigma']
        mu = 0.5 * (np.log((lambD / 10)**2/(np.exp(sigma**2) - 1)) - sigma**2)
        mu_sign = np.sign(mu)
        exp = np.exp(mu - sigma**2)

        # sigma
        kvar = 'sigma'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            x0[ival] = 1

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=scales,
                )

        # mu
        kvar = 'mu'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            x0[ival] = mu_sign

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=scales,
                )

        # t0
        # max at lamb - (lamb00 + lambD * tau) = exp(mu - sigma**2)
        kvar = 'tau'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            x0[ival] = (lamb_ext - exp - lamb0) / lambD / scales[ival]

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=scales,
                )

        # amp
        # max = amp * exp(sigma**2/2 - mu)
        kvar = 'amp'
        vind = dind['jac'][kfunc].get(kvar)
        if vind is not None:
            ival, _ = vind['val'], vind['var']
            x0[ival] = data_pulse_sign

            for ii, (din, val) in enumerate(ldins):
                _update_din_from_user(
                    din, kfunc, kvar, val,
                    scales=scales,
                )

    return x0


# #####################################################################
# #####################################################################
#                     Utilities
# #####################################################################


def _prepare(
    lamb=None,
    data=None,
    iok=None,
    axis=None,
    # binning
    dbinning=None,
):

    # ---------------
    # lamb scales
    # ----------------

    lamb0 = lamb[0]
    lambD = np.abs(lamb[-1] - lamb[0])
    lambm = np.mean(lamb)

    # ------------------------
    # indices of min max data
    # ------------------------

    idatamax = np.nanargmax(data[iok])
    idatamin = np.nanargmin(data[iok])

    # -------------------------
    # shapes adjustment
    # -------------------------

    if dbinning is not False:
        bin_dlamb = dbinning['dlamb'][dbinning['ind']]

    # time dependence adjustments
    if lamb.shape != iok.shape:
        lambn = np.full(iok.shape, np.nan)
        sli = [0 if ii == axis else slice(None) for ii in range(iok.ndim)]
        for ii, ll in enumerate(lamb):
            sli[axis] = ii
            lambn[tuple(sli)] = ll
        lamb = lambn

        if dbinning is not False:
            bin_dlambn = np.full(iok.shape, np.nan)
            sli = [0 if ii == axis else slice(None) for ii in range(iok.ndim)]
            for ii, ll in enumerate(bin_dlamb):
                sli[axis] = ii
                bin_dlambn[tuple(sli)] = ll
            bin_dlamb = bin_dlambn

    # --------------
    # lamb min max
    # --------------

    lamb_amax = lamb[iok][idatamax]
    lamb_amin = lamb[iok][idatamin]

    # --------------
    # data min max
    # --------------

    if dbinning is False:

        data_max = data[iok][idatamax]
        data_min = data[iok][idatamin]

        data_mean = np.nanmean(data[iok])
        data_median = np.median(data[iok])

    else:

        nbins = dbinning['binning']
        data_max = data[iok][idatamax] / (bin_dlamb[iok][idatamax] * nbins)
        data_min = data[iok][idatamin] / (bin_dlamb[iok][idatamin] * nbins)

        data_mean = np.nanmean(data[iok]) / (bin_dlamb[iok][idatamin] * nbins)
        data_median = np.median(data[iok]) / (bin_dlamb[iok][idatamin] * nbins)

    # -------------------
    # sign and extrema
    # -------------------

    data_pulse_sign = np.sign(data_mean - data_median)
    lamb_ext = lamb_amax if data_pulse_sign > 0 else lamb_amin

    return (
        lamb0, lambD, lambm,
        lamb_amax, lamb_amin,
        lamb_ext,
        data_max, data_min,
        data_mean, data_median,
        data_pulse_sign,
    )


# #######################
# generic pdate from dict
# #######################


def _update_din_from_user(din, kfunc, kvar, val, scales=None):

    if din.get(kfunc, {}).get(kvar) is not None:
        ind = din[kfunc][kvar]['ind']
        if scales is None:
            val[ind] = din[kfunc][kvar]['val']
        else:
            val[ind] = din[kfunc][kvar]['val'] / scales[ind]

    return


# ###########################
# generic pulse scales bounds
# ###########################


def _get_scales_bounds_pulse(
    dind=None,
    kfunc=None,
    ldins=None,
    # data
    data_min=None,
    data_max=None,
    data_pulse_sign=None,
    # lamb
    lamb=None,
    lambm=None,
    lambD=None,
    # arrays to fill
    scales=None,
    bounds0=None,
    bounds1=None,
):

    # amp
    kvar = 'amp'
    vind = dind['jac'][kfunc].get(kvar)
    if vind is not None:
        ival, _ = vind['val'], vind['var']
        scales[ival] = (data_max - data_min)
        if data_pulse_sign > 0.:
            bounds0[ival] = 0
            bounds1[ival] = 10.
        else:
            bounds0[ival] = -10.
            bounds1[ival] = 0.

        for ii, (din, val) in enumerate(ldins):
            _update_din_from_user(
                din, kfunc, kvar, val,
                scales=None if ii == 0 else scales
            )

    # tau
    kvar = 'tau'
    vind = dind['jac'][kfunc].get(kvar)
    if vind is not None:
        ival, _ = vind['val'], vind['var']
        scales[ival] = 1
        bounds0[ival] = 0
        bounds1[ival] = 1

        for ii, (din, val) in enumerate(ldins):
            _update_din_from_user(
                din, kfunc, kvar, val,
                scales=None if ii == 0 else scales
            )

    # tup
    kvar = 't_up'
    vind = dind['jac'][kfunc].get(kvar)
    if vind is not None:
        ival, _ = vind['val'], vind['var']
        scales[ival] = 0.05 * lambD
        bounds0[ival] = 1.e-3
        bounds1[ival] = 20

        for ii, (din, val) in enumerate(ldins):
            _update_din_from_user(
                din, kfunc, kvar, val,
                scales=None if ii == 0 else scales
            )

    # tdown
    kvar = 't_down'
    vind = dind['jac'][kfunc].get(kvar)
    if vind is not None:
        ival, _ = vind['val'], vind['var']
        scales[ival] = 0.2 * lambD
        bounds0[ival] = 1.e-3
        bounds1[ival] = 20

        for ii, (din, val) in enumerate(ldins):
            _update_din_from_user(
                din, kfunc, kvar, val,
                scales=None if ii == 0 else scales
            )

    return


# #######################
# generic pulse x0
# #######################


def _get_x0_pulse(
    dind=None,
    kfunc=None,
    ldins=None,
    # data
    data_min=None,
    data_max=None,
    data_pulse_sign=None,
    # lamb
    lambm=None,
    lambD=None,
    lamb0=None,
    lamb_ext=None,
    # arrays to fill
    x0=None,
    scales=None,
):

    # amp
    kvar = 'amp'
    vind = dind['jac'][kfunc].get(kvar)
    if vind is not None:
        ival, _ = vind['val'], vind['var']
        x0[ival] = (data_max - data_min) * data_pulse_sign / scales[ival]

        for ii, (din, val) in enumerate(ldins):
            _update_din_from_user(
                din, kfunc, kvar, val,
                scales=scales,
            )

    # tau
    kvar = 'tau'
    vind = dind['jac'][kfunc].get(kvar)
    if vind is not None:
        ival, _ = vind['val'], vind['var']
        x0[ival] = (lamb_ext - lamb0) / lambD

        for ii, (din, val) in enumerate(ldins):
            _update_din_from_user(
                din, kfunc, kvar, val,
                scales=scales,
            )

    # tup
    kvar = 't_up'
    vind = dind['jac'][kfunc].get(kvar)
    if vind is not None:
        ival, _ = vind['val'], vind['var']
        x0[ival] = 1

        for ii, (din, val) in enumerate(ldins):
            _update_din_from_user(
                din, kfunc, kvar, val,
                scales=scales,
            )

    # tdown
    kvar = 't_down'
    vind = dind['jac'][kfunc].get(kvar)
    if vind is not None:
        ival, _ = vind['val'], vind['var']
        x0[ival] = 1

        for ii, (din, val) in enumerate(ldins):
            _update_din_from_user(
                din, kfunc, kvar, val,
                scales=scales,
            )

    return


# -*- coding: utf-8 -*-
"""
Created on Tue Jun 18 09:33:27 2024

@author: dvezinet
"""


import datetime as dtm
import itertools as itt


import numpy as np
import datastock as ds


#############################################
#############################################
#       main
#############################################


def main(
    coll=None,
    key_model=None,
    key_data=None,
    lamb=None,
    # options
    details=None,
    binning=None,
    # uncertainty propagation
    uncertainty_method=None,
    # others
    returnas=None,
    store=None,
    store_key=None,
    # timing
    timing=None,
):

    # -----------------
    # check
    # -----------------

    (
        key_model, ref_nx, ref_nf,
        key_data, key_cov, axis,
        key_lamb, lamb, ref_lamb,
        binning, details,
        uncertainty_method,
        returnas, store, store_key,
        timing,
    ) = _check(
        coll=coll,
        key_model=key_model,
        key_data=key_data,
        lamb=lamb,
        # others
        binning=binning,
        details=details,
        # uncertainty propagation
        uncertainty_method=uncertainty_method,
        # others
        returnas=returnas,
        store=store,
        store_key=store_key,
        # timing
        timing=timing,
    )

    if timing is True:
        fname = 'interpolate_spectral_model()'
        t0 = dtm.datetime.now()  # DB

    # -----------------
    # optionnal binning
    # -----------------

    dbinning = coll.get_spectral_fit_binning_dict(
        binning=binning,
        lamb=lamb,
        iok=None,
        axis=None,
    )

    if timing is True:
        t1 = dtm.datetime.now()  # DB
        print(f'... timing {fname}: binning dict {(t1-t0).total_seconds()} s')

    # ----------------
    # prepare
    # ----------------

    # ----------
    # data_in

    data_in = coll.ddata[key_data]['data']
    ref_in = coll.ddata[key_data]['ref']
    ndim_in = data_in.ndim

    # ----------
    # details

    iref_nx = ref_in.index(ref_nx)
    if details is True:
        iref_nx_out = iref_nx + 1
    else:
        iref_nx_out = iref_nx

    # ----------
    # coraviance

    if key_cov is not None:
        cov = coll.ddata[key_cov]['data']
        nx = coll.dref[ref_nx]['size']

    if timing is True:
        t2 = dtm.datetime.now()  # DB
        print(f'... timing {fname}: prepare data {(t2-t1).total_seconds()} s')

    # -----------------------
    # prepare loop on indices
    # -----------------------

    key_bs = None
    if key_bs is None:
        lind = [
            range(ss) for ii, ss in enumerate(data_in.shape)
            if ii != iref_nx
        ]
    else:
        raise NotImplementedError()

    # -------------
    # initialize
    # -------------

    # shape_out, ref_out
    shape_in = data_in.shape
    shape_out = list(shape_in)
    ref_out = list(ref_in)

    shape_out[iref_nx] = lamb.size
    ref_out[iref_nx] = ref_lamb
    if details is True:
        shape_out.insert(0, coll.dref[ref_nf]['size'])
        ref_out.insert(0, ref_nf)

    # data_out
    data_out = np.full(tuple(shape_out), np.nan)

    if timing is True:
        t3 = dtm.datetime.now()  # DB
        print(f'... timing {fname}: initialize {(t3-t2).total_seconds()} s')

    # ----------------
    # get func
    # ----------------

    if details is True:
        func = coll.get_spectral_fit_func(
            key=key_model,
            func='details',
        )['details']

    else:
        func = coll.get_spectral_fit_func(
            key=key_model,
            func='sum',
        )['sum']

    if timing is True:
        t4 = dtm.datetime.now()  # DB
        print(f'... timing {fname}: get func {(t4-t3).total_seconds()} s')

    # --------------
    # prepare slices
    # --------------

    # slices
    sli_in = list(shape_in)
    sli_out = list(shape_out)

    sli_in[iref_nx] = slice(None)
    sli_out[iref_nx_out] = slice(None)
    if details is True:
        sli_out[0] = slice(None)

    # as array
    sli_in = np.array(sli_in)
    sli_out = np.array(sli_out)

    # indices to change
    if ndim_in > 1:
        ind0 = np.array(
            [ii for ii in range(len(shape_in)) if ii != iref_nx],
            dtype=int,
        )

        # adjust for details
        if details is True:
            ind0_out = ind0 + 1
        else:
            ind0_out = ind0

    else:
        ind0 = None

    if timing is True:
        t5 = dtm.datetime.now()  # DB
        print(f'... timing {fname}: prepare slices {(t5-t4).total_seconds()} s')

    # -------------------------
    # loop to compute data_out
    # -------------------------

    if dbinning is False:
        bin_ind = False
        bin_dlamb = None
    else:
        bin_ind = dbinning['ind']
        bin_dlamb = dbinning['dlamb']

    for ind in itt.product(*lind):

        # update slices
        if ind0 is not None:
            sli_in[ind0] = ind
            sli_out[ind0_out] = ind

        # call func
        data_out[tuple(sli_out)] = func(
            x_free=data_in[tuple(sli_in)],
            lamb=lamb if dbinning is False else dbinning['lamb'],
            bin_ind=bin_ind,
            bin_dlamb=bin_dlamb,
        )

    if timing is True:
        t6 = dtm.datetime.now()  # DB
        print(f'... timing {fname}: data_out {(t6-t5).total_seconds()} s')

    # -----------------------------
    # loop on cov to get error bar
    # -----------------------------

    if key_cov is not None:

        data_min, data_max = _uncertainty_propagation(
            # resources
            coll=coll,
            key_model=key_model,
            lamb=lamb,
            data_in=data_in,
            data_out=data_out,
            axis=axis,
            nx=nx,
            lind=lind,
            func=func,
            # method
            method=uncertainty_method,
            # slicing
            ind0=ind0,
            ind0_out=ind0_out,
            sli_in=sli_in,
            sli_out=sli_out,
            # uncertainty input
            cov=cov,
            # binning
            dbinning=dbinning,
            bin_ind=bin_ind,
            bin_dlamb=bin_dlamb,
            # timing
            timing=timing,
        )

        if timing is True:
            t7 = dtm.datetime.now()  # DB
            print(f'... timing {fname}: uncertainty {(t7-t6).total_seconds()} s')

    else:
        data_min = None
        data_max = None

    # --------------
    # return
    # --------------

    dout = {
        'key': store_key,
        'key_data': key_data,
        'key_model': key_model,
        'key_lamb': key_lamb,
        'lamb': lamb,
        'details': details,
        'data': data_out,
        'data_min': data_min,
        'data_max': data_max,
        'ref': tuple(ref_out),
        'dim': coll.ddata[key_data]['dim'],
        'quant': coll.ddata[key_data]['quant'],
        'units': coll.ddata[key_data]['units'],
    }

    # --------------
    # store
    # --------------

    if store is True:

        lout = [
            'key_data', 'key_model', 'key_lamb',
            'lamb', 'details',
            'data_min', 'data_max',
        ]
        coll.add_data(
            **{
                k0: v0 for k0, v0 in dout.items()
                if k0 not in lout
            },
        )

    return dout


#############################################
#############################################
#       check
#############################################


def _check(
    coll=None,
    key_model=None,
    key_data=None,
    lamb=None,
    # others
    binning=None,
    details=None,
    # uncertainty propagation
    uncertainty_method=None,
    # others
    returnas=None,
    store=None,
    store_key=None,
    # timing
    timing=None,
):

    # ---------------------
    # key_model, key_data
    # ---------------------

    key_model, key_data, key_cov, lamb, binning = _check_keys(
        coll=coll,
        key=key_model,
        key_data=key_data,
        lamb=lamb,
        binning=binning,
    )

    # derive ref_model
    wsm = coll._which_model
    ref_nf = coll.dobj[wsm][key_model]['ref_nf']
    ref_nx = coll.dobj[wsm][key_model]['ref_nx']

    axis = coll.ddata[key_data]['ref'].index(ref_nx)

    # -----------------
    # lamb
    # -----------------

    if isinstance(lamb, np.ndarray):
        c0 = (
            lamb.ndim == 1
            and np.all(np.isfinite(lamb))
        )
        if not c0:
            _err_lamb(lamb)

        key_lamb = None
        ref_lamb = None

    elif isinstance(lamb, str):
        c0 = (
            lamb in coll.ddata.keys()
            and coll.ddata[lamb]['data'].ndim == 1
            and np.all(np.isfinite(coll.ddata[lamb]['data']))
        )
        if not c0:
            _err_lamb(lamb)

        key_lamb = lamb
        lamb = coll.ddata[key_lamb]['data']
        ref_lamb = coll.ddata[key_lamb]['ref'][0]

    else:
        _err_lamb(lamb)

    # -----------------
    # details
    # -----------------

    details = ds._generic_check._check_var(
        details, 'details',
        types=bool,
        default=False,
    )

    # -----------------
    # uncertainty_method
    # -----------------

    uncertainty_method = ds._generic_check._check_var(
        uncertainty_method, 'uncertainty_method',
        types=str,
        default='standard',
        allowed=['standard', 'min/max'],
    )

    # -----------------
    # store
    # -----------------

    lok = [False]
    if key_lamb is not None:
        lok.append(True)
    store = ds._generic_check._check_var(
        store, 'store',
        types=bool,
        default=False,
        allowed=lok,
    )

    # -----------------
    # returnas
    # -----------------

    returnas = ds._generic_check._check_var(
        returnas, 'returnas',
        default=(not store),
        allowed=[False, True, dict],
    )

    # -----------------
    # store_key
    # -----------------

    if store is True:
        lout = list(coll.ddata.keys())
        store_key = ds._generic_check._check_var(
            store_key, 'store_key',
            types=str,
            excluded=lout,
        )
    else:
        store_key = None

    # -----------------
    # timing
    # -----------------

    timing = ds._generic_check._check_var(
        timing, 'timing',
        types=bool,
        default=False,
    )

    return (
        key_model, ref_nx, ref_nf,
        key_data, key_cov, axis,
        key_lamb, lamb, ref_lamb,
        binning, details,
        uncertainty_method,
        returnas, store, store_key,
        timing,
    )


def _check_keys(coll=None, key=None, key_data=None, lamb=None, binning=None):

    # ---------------------
    # key_model vs key_fit
    # ---------------------

    wsm = coll._which_model
    wsf = coll._which_fit

    lokm = list(coll.dobj.get(wsm, {}).keys())
    lokf = list(coll.dobj.get(wsf, {}).keys())

    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lokm + lokf,
    )

    # ---------------
    # if key_fit
    # ---------------

    key_cov = None
    if key in lokf:
        key_fit = key
        key_model = coll.dobj[wsf][key_fit]['key_model']

        if key_data is None:
            key_data = coll.dobj[wsf][key_fit]['key_sol']
            key_cov = coll.dobj[wsf][key_fit]['key_cov']

        if lamb is None:
            lamb = coll.dobj[wsf][key_fit]['key_lamb']

        binning = coll.dobj[wsf][key_fit]['dinternal']['binning']

    else:
        key_model = key

    # ----------
    # key_data
    # ----------

    # derive ref_model
    ref_nx = coll.dobj[wsm][key_model]['ref_nx']

    # list of acceptable values
    lok = [
        k0 for k0, v0 in coll.ddata.items()
        if ref_nx in v0['ref']
    ]

    # check
    key_data = ds._generic_check._check_var(
        key_data, 'key_data',
        types=str,
        allowed=lok,
    )

    return key_model, key_data, key_cov, lamb, binning


def _err_lamb(lamb):
    msg = (
        "Arg 'lamb' nust be either:\n"
        "\t- 1d np.ndarray with finite values only\n"
        "\t- str: a key to an existing 1d vector with finite values only\n"
        f"Provided:\n{lamb}"
    )
    raise Exception(msg)


#############################################
#############################################
#       uncertainty propagation
#############################################


def _uncertainty_propagation(
    # resources
    coll=None,
    key_model=None,
    lamb=None,
    data_in=None,
    data_out=None,
    axis=None,
    nx=None,
    lind=None,
    func=None,
    # method
    method=None,
    # slicing
    ind0=None,
    ind0_out=None,
    sli_in=None,
    sli_out=None,
    # uncertainty input
    cov=None,
    # binning
    dbinning=None,
    bin_ind=None,
    bin_dlamb=None,
    # timing
    timing=None,
):

    # ------------------
    # prepare
    # ------------------

    if timing is True:
        i0 = 0  # DB
        dt0 = 0
        dt1 = 0

    # prepare output
    data_min = np.full(data_out.shape, np.inf)
    data_max = np.full(data_out.shape, -np.inf)

    # sli_cov
    sli_cov = np.r_[
        [0 for ii in range(data_in.ndim-1)] + [slice(None), slice(None)]
    ]
    ind_cov = np.arange(data_in.ndim-1)

    sli_cov = [0 for ii in cov.shape]
    sli_cov[axis] = slice(None)
    sli_cov[axis+1] = slice(None)
    sli_cov = np.array(sli_cov)

    ind_cov = np.array(
        [ii for ii in range(len(cov.shape)) if ii not in [axis, axis+1]],
        dtype=int,
    )

    # -------------------------------------------------------
    # Proper uncertainty propagation using covariance matrix
    # -------------------------------------------------------

    # ref: https://en.wikipedia.org/wiki/Propagation_of_uncertainty

    if method == 'standard':

        # ------------------------
        # get jacobian at solution

        # function
        func_jac = coll.get_spectral_fit_func(
            key=key_model,
            func='jac',
        )['jac']

        # ------------------------
        # loop on all spectra

        scales = np.ones((nx,), dtype=float)
        for ind in itt.product(*lind):

            # update slices
            if ind0 is not None:
                sli_in[ind0] = ind
                sli_out[ind0_out] = ind
                sli_cov[ind_cov] = ind

            datain = data_in[tuple(sli_in)]

            # values
            jac = func_jac(
                x_free=datain,
                lamb=lamb if dbinning is False else dbinning['lamb'],
                bin_ind=bin_ind,
                bin_dlamb=bin_dlamb,
                scales=scales,
            )

            # derive std of sum
            cov_sum = np.sqrt((jac).dot(cov[tuple(sli_cov)].dot(jac.T)))
            std_sum = np.sqrt(np.diag(cov_sum))

            # ------------------------
            # derive min, max

            sliout = tuple(sli_out)
            data_min[sliout] = data_out[sliout] - std_sum
            data_max[sliout] = data_out[sliout] + std_sum

    # ------------------------------------------------------
    # brute force uncertainty propagation using min / max of all variables
    # ------------------------------------------------------

    else:

        # prepare
        inc = np.r_[-1, 0, 1]
        lind_std = [inc for ii in range(nx)]

        # -------------------
        # loop on spectra

        for ind in itt.product(*lind):

            # update slices
            if ind0 is not None:
                sli_in[ind0] = ind
                sli_out[ind0_out] = ind
                sli_cov[ind_cov] = ind

            # std
            std = np.sqrt(np.diag(cov[tuple(sli_cov)]))

            # loop on [-1, 0, 1] combinations
            for stdi in itt.product(*lind_std):

                # data = data_in + std * (-1, 0, 1)
                datain = (
                    data_in[tuple(sli_in)]
                    + np.r_[stdi] * std
                )

                if timing is True:
                    t00 = dtm.datetime.now()

                # call func
                datai = func(
                    x_free=datain,
                    lamb=lamb if dbinning is False else dbinning['lamb'],
                    bin_ind=bin_ind,
                    bin_dlamb=bin_dlamb,
                )

                if timing is True:
                    t01 = dtm.datetime.now()
                    dt0 += (t01 - t00).total_seconds()

                # update min, max
                data_min[tuple(sli_out)] = np.minimum(
                    data_min[tuple(sli_out)], datai,
                )
                data_max[tuple(sli_out)] = np.maximum(
                    data_max[tuple(sli_out)],
                    datai,
                )

                if timing is True:
                    t02 = dtm.datetime.now()
                    dt1 += (t02 - t01).total_seconds()
                    i0 += 1

        if timing is True:
            msg = (
                f"\t... timing std: func {dt0} s\n"
                f"\t... timing std: max/min {dt1} s\n"
                f"\t... timing std: ntot = {i0}\n"
            )
            print(msg)

    return data_min, data_max


# -*- coding: utf-8 -*-
"""
Created on Wed Jul 24 11:06:21 2024

@author: dvezinet
"""


import numpy as np


#############################################
#############################################
#       MODEL FUNC ORDER
#############################################


_LMODEL_ORDER = [
    # background
    'poly', 'exp_lamb', # 'exp_E',
    # spectral lines
    'gauss', 'lorentz', 'pvoigt', 'voigt',
    # pulse shape
    'pulse_exp', 'pulse_gauss', 'lognorm',
]


#############################################
#############################################
#       MODEL DICT
#############################################


_DMODEL = {

    # -------------------
    # background
    # -------------------

    # -----------
    # poly

    'poly': {
        'var': ['a0', 'a1', 'a2'],
        'description': 'polynomial (up to deg = 2)',
        'expressions': {
            'main': r"$\left\{ \begin{array}{ll} a_0 + a_1\frac{\lambda - <\lambda>}{\Delta\lambda} + a_2\left(\frac{\lambda - <\lambda>}{\Delta\lambda}\right)^2\\ <\lambda> = \frac{\lambda[0] + \lambda[-1]}{2} \\ \Delta\lambda = \lambda[-1] - \lambda[0]\end{array} \right.$",
            'argmax': r"$\lambda_{max} = <\lambda> - \frac{a_1}{2a_2}\Delta\lambda$",
            'max': r"$f(\lambda_{max}) = a_0 - \frac{a_1^2}{4a_2}$",
        },
    },

    # ------------
    # exponential

    'exp_lamb': {
        'var': ['amp', 'rate'],
        'description': 'Bremsstrahlung-like exponential',
        'expressions': {
            'main': r"$\frac{A}{\lambda} e^{-\frac{rate}{\lambda}}$",
            'Te': r"$k_BT_e [eV] = \frac{hc}{rate} \times \frac{1}{e}$",
        },
    },

    # -------------------
    # spectral lines
    # -------------------

    # -----------
    # gaussian

    'gauss': {
        'var': ['amp', 'vccos', 'sigma'],
        'param': [('lamb0', float), ('mz', float, np.nan)],
        'description': 'gaussian',
        'expressions': {
            'main': r"$A e^{-\frac{\left(\lambda - \lambda_0(1 + \frac{v}{c}\cos)\right)^2}{2\sigma^2}}$",
            'argmax': r"$\lambda_{max} = \lambda_0(1 + \frac{v}{c}\cos)$",
            'max': r"$f(\lambda_{max}) = A$",
            'Ti': r"$k_BT_i [eV] = m_z c^2 \left(\frac{\sigma}{\lambda_0}\right)^2 \times \frac{1}{e}$",
            "FWHM": "",
        },
    },

    # -----------
    # lorentzian

    'lorentz': {
        'var': ['amp', 'vccos', 'gam'],
        'param': [('lamb0', float), ('mz', float, np.nan)],
        'description': 'lorentzian',
        'expressions': {
            'main': r"$\frac{A}{1 + \left(\frac{\lambda - \lambda_0(1 + \frac{v}{c}\cos)}{\gamma}\right)^2}$",
            'argmax': r"$\lambda_{max} = \lambda_0(1 + \frac{v}{c}\cos)$",
            'max': r"$f(\lambda_{max}) = A$",
        },
    },

    # ------------
    # pseudo-voigt

    'pvoigt': {
        'var': ['amp', 'vccos', 'sigma', 'gam'],
        'param': [('lamb0', float), ('mz', float, np.nan)],
        'description': 'pseudo-voigt',
        'expressions': {
            'main': (
                r"$\left\{ \begin{array}{ll} A \left[\frac{\eta}{1 + \left(\frac{\lambda - \lambda_0(1 + \frac{v}{c}\cos)}{\gamma}\right)^2} + (1-\eta)e^{-\frac{\left(\lambda - \lambda_0(1 + \frac{v}{c}\cos)\right)^2}{2\sigma^2}}\right] \\ \eta = g(\sigma, \gamma) \end{array} \right.$"
            ),
            'ref': "https://en.wikipedia.org/wiki/Voigt_profile",
            'argmax': r"$\lambda_{max} = \lambda_0(1 + \frac{v}{c}\cos)$",
            'max': r"$f(\lambda_{max}) = A$",
            'Ti': r"$k_BT_i [eV] = m_z c^2 \left(\frac{\sigma}{\lambda_0}\right)^2 \times \frac{1}{e}$",
        },
    },

    # -----------
    # voigt

    'voigt': {
        'var': ['amp', 'vccos', 'sigma', 'gam'],
        'param': [('lamb0', float), ('mz', float, np.nan)],
        'description': 'voigt',
        'expressions': {
            'main': r"$$",
            'argmax': "",
            'max': "",
            'Ti': r"$k_BT_i [eV] = m_z c^2 \left(\frac{\sigma}{\lambda_0}\right)^2 \times \frac{1}{e}$",
        },
    },

    # ------------------
    # pulse shape
    # ------------------

    # -----------
    # pulse with exponentials

    'pulse_exp': {
        'var': ['amp', 'tau', 't_up', 't_down'],
        'description': 'asymmetric pulse, 2 exponentials',
        'expressions': {
            'main': (
                r"$\left\{ \begin{array}{ll} f(t>= t_0) = A\left( e^{-\frac{t-t_0}{t_{down}}} - e^{-\frac{t-t_0}{t_{up}}} \right) \\ t_0 = t[0] + \tau\Delta t  \\ \Delta t = t[-1] - t[0] \end{array} \right.$"
            ),
            'argmax': r"$t_{max} = t_0 + \frac{t_{down}t_{up}}{t_{down} - t_{up}}\ln\left({\frac{t_{down}}{t_{up}}}\right)$",
            'max': r"$f(t_{max}) = A\left( e^{-\frac{t_{up}}{t_{down} - t_{up}}\ln\left({\frac{t_{down}}{t_{up}}}\right)} - e^{-\frac{t_{down}}{t_{down} - t_{up}}\ln\left({\frac{t_{down}}{t_{up}}}\right)}\right)$",
        },
    },

    # -----------
    # pulse with gaussians

    'pulse_gauss': {
        'var': ['amp', 'tau', 't_up', 't_down'],
        'description': 'asymmetric pulse, 2 gaussians',
        'expressions': {
            'main': (
                r"$\left\{ \begin{array}{ll} f(t < t_0) = A e^{-\left(\frac{t-t_0}{t_{up}}\right)^2} \\ f(t>=t_0) = A e^{-\left(\frac{t-t_0}{t_{down}}\right)^2} \\ t_0 = t[0] + \tau\Delta t  \\ \Delta t = t[-1] - t[0] \end{array} \right.$"
            ),
            'argmax': r"$t_{max} = t_0$",
            'max': r"$f(t_{max}) = A$",
        },
    },

    # -----------
    # lognorm

    'lognorm': {
        'var': ['amp', 'tau', 'mu', 'sigma'],
        'description': 'asymmetric pulse, lognorm',
        'expressions': {
            'main': r"$\left\{ \begin{array}{ll} f(t > t_0) = \frac{A}{t-t_0} e^{-\frac{\left(\ln(t-t_0) - \mu\right)^2}{2\sigma^2}} \\ t_0 = t[0] + \tau\Delta t  \\ \Delta t = t[-1] - t[0] \end{array} \right.$",
            'argmax': r"$t_{max} = t_0 + e^{\mu - \sigma^2}$",
            'max': r"$f(t_{max}) = A e^{\frac{\sigma^2}{2} - \mu}$",
            'skewness': r"$skew = \left(e^{\sigma^2} + 2\right) \sqrt{exp(\sigma^2) - 1}$",
            'variance': r"$var = \left(e^{\sigma^2} - 1\right) e^{2\mu + \sigma^2}$"
        },
    },
}


#!/usr/bin/env python

# Built-in
import sys
import os
import argparse


_HERE = os.path.abspath(os.path.dirname(__file__))


# import parser dict
sys.path.insert(1, _HERE)
from _dparser import _DPARSER
_ = sys.path.pop(1)


###################################################
###################################################
#       default values
###################################################


_PKG = 'spectrally'
_PKGPATH = os.path.dirname(_HERE)
_ENTRYPOINTS_PATH = os.path.join(_PKGPATH, _PKG, 'entrypoints')


_LOPTIONS = ['--version', 'custom']
_LOPSTRIP = [ss.strip('--') for ss in _LOPTIONS]


###################################################
###################################################
#       function
###################################################


def spectrally_bash(option=None, ddef=None, **kwdargs):
    f""" Print {_PKG} version and / or store in environment variable """

    # --------------
    # Check inputs
    # --------------

    if option not in _LOPSTRIP:
        msg = (
            "Provided option is not acceptable:\n"
            f"\t- available: {_LOPSTRIP}\n"
            f"\t- provided:  {option}"
        )
        raise Exception(msg)

    # --------------------------------
    # call corresponding bash command
    # --------------------------------

    # version
    if option == 'version':
        sys.path.insert(1, _HERE)
        import spectrallyversion
        _ = sys.path.pop(1)
        spectrallyversion.get_version(ddef=ddef, **kwdargs)

    # custom
    elif option == 'custom':
        sys.path.insert(1, _HERE)
        import spectrallycustom
        _ = sys.path.pop(1)
        spectrallycustom.custom(ddef=ddef, **kwdargs)

    return


###################################################
###################################################
#       bash call (main)
###################################################


def main():

    # ----------------------
    # Parse input arguments
    # ----------------------

    # description msg
    msg = (
        f"Get {_PKG} version from bash\n\n"
        "Optionally set an enviroment variable\n"
        "If run from a git repo containing {_PKG}, returns git describe\n"
        "Otherwise reads the {_PKG} version stored in {_PKG}/version.py\n"
    )

    # Instanciate parser
    parser = argparse.ArgumentParser(description=msg)

    # ----------------------
    # Define input arguments
    # ----------------------

    # option
    parser.add_argument(
        'option',
        nargs='?',
        type=str,
        default='None',
    )

    # version
    parser.add_argument(
        '-v', '--version',
        help=f'get {_PKG} current version',
        required=False,
        action='store_true',
    )

    # kwd
    parser.add_argument(
        'kwd',
        nargs='?',
        type=str,
        default='None',
    )

    # ----------------------
    # Options
    # ----------------------

    # check provided options
    if sys.argv[1] not in _LOPTIONS:
        msg = (
            "Provided option is not acceptable:\n"
            f"\t- available: {_LOPTIONS}\n"
            f"\t- provided:  {sys.argv[1]}"
        )
        raise Exception(msg)

    # check nb of options
    if len(sys.argv) > 2:
        if any([ss in sys.argv[2:] for ss in _LOPTIONS]):
            lopt = [ss for ss in sys.argv[1:] if ss in _LOPTIONS]
            msg = (
                "Only one option can be provided!\n"
                f"\t- provided: {lopt}"
            )
            raise Exception(msg)

    # get option
    option = sys.argv[1].strip('--')

    # parse to get kwdargs
    ddef, parser = _DPARSER[option]()
    if len(sys.argv) > 2:
        kwdargs = dict(parser.parse_args(sys.argv[2:])._get_kwargs())
    else:
        kwdargs = {}

    # ----------------------
    # call function
    # ----------------------

    spectrally_bash(option=option, ddef=ddef, **kwdargs)

    return


###################################################
###################################################
#       __main__
###################################################


if __name__ == '__main__':
    main()


# #!/usr/bin/python3
# -*- coding: utf-8 -*-


import numpy as np
import matplotlib.pyplot as plt
from matplotlib import gridspec
import datastock as ds


# ###############################################################
# ###############################################################
#               Main
# ###############################################################


def main(
    coll=None,
    key_model=None,
    key_data=None,
    lamb=None,
    keyY=None,
    # options
    details=None,
    # plotting
    dprop=None,
    vmin=None,
    vmax=None,
    # figure
    dax=None,
    fs=None,
    dmargin=None,
    tit=None,
    connect=None,
    dinc=None,
    show_commands=None,
):

    # -------------------
    # check
    # -------------------

    details, connect = _check(
        details=details,
        connect=connect,
    )

    # -------------------
    # interpolate
    # -------------------

    dout = coll.interpolate_spectral_model(
        key_model=key_model,
        key_data=key_data,
        lamb=lamb,
        # options
        details=details,
        # others
        returnas=dict,
    )

    # -------------------
    # extract coll2
    # -------------------

    coll2, dkeys, ndim = _extract_coll2(
        coll=coll,
        key_model=key_model,
        dout=dout,
        details=details,
        keyY=keyY,
    )

    # -------------------
    # prepare figure
    # -------------------

    if dax is None:
        dax = _get_dax(
            ndim=ndim,
            fs=fs,
            dmargin=dmargin,
            tit=tit,
        )

    dax = ds._generic_check._check_dax(dax)

    # -------------------
    # plot
    # -------------------

    if ndim == 1:
        dax = _plot_1d(
            coll2,
            dout=dout,
            dkeys=dkeys,
            dax=dax,
            details=details
        )

    elif ndim == 2:
        dax, dgroup = _plot_2d(
            coll=coll,
            key_model=key_model,
            coll2=coll2,
            dout=dout,
            keyY=keyY,
            dkeys=dkeys,
            dax=dax,
            details=details,
        )

    # -------------------
    # finalize
    # -------------------

    _finalize_figure(
        dax=dax,
        dout=dout,
        tit=tit,
    )

    # ---------------------
    # connect interactivity
    # ---------------------

    if isinstance(dax, dict):
        return dax
    else:
        if connect is True:
            dax.setup_interactivity(kinter='inter0', dgroup=dgroup, dinc=dinc)
            dax.disconnect_old()
            dax.connect()

            dax.show_commands(verb=show_commands)
            return dax
        else:
            return dax, dgroup


# ###############################################################
# ###############################################################
#               check
# ###############################################################


def _check(
    details=None,
    connect=None,
):

    # -------------
    # details
    # -------------

    details = ds._generic_check._check_var(
        details, 'details',
        types=bool,
        default=True,
    )

    # -------------
    # connect
    # -------------

    connect = ds._generic_check._check_var(
        connect, 'connect',
        types=bool,
        default=True,
    )

    return details, connect


# ###############################################################
# ###############################################################
#               extract coll2
# ###############################################################


def _extract_coll2(
    coll=None,
    key_model=None,
    dout=None,
    details=None,
    keyY=None,
):

    # ------------
    # initialize
    # -------------

    wsm = coll._which_model
    key = 'data'
    key_lamb = dout['key_lamb']
    if key_lamb is None:
        key_lamb = 'lamb'

    coll2 = coll.__class__()

    # --------------
    # add all refs
    # --------------

    # ref_nfunc
    ref = dout['ref']
    if details is True:
        ref = list(ref)
        ref[0] = 'nfunc'
        coll2.add_ref(ref[0], size=dout['data'].shape[0])

    # kre_lamb
    if dout['key_lamb'] is None:
        ilamb = ref.index(None)
        kref_lamb = 'nlamb'
        ref = list(ref)
        ref[ilamb] = kref_lamb
        coll2.add_ref(kref_lamb, size=dout['data'].shape[ilamb])

    # all other refs
    for rr in ref:
        if rr not in coll2.dref.keys():
            coll2.add_ref(key=rr, size=coll.dref[rr]['size'])

    # --------------
    # add all data
    # --------------

    lk = ['data', 'units', 'dim', 'quant']

    # -------
    # lamb

    if dout['key_lamb'] is None:
        coll2.add_data(key_lamb, data=dout['lamb'], ref=kref_lamb)
    else:
        coll2.add_data(
            dout['key_lamb'],
            **{k0: coll.ddata[dout['key_lamb']][k0] for k0 in lk + ['ref']},
        )

    # ---------------
    # sum + details

    if details is True:

        # data
        lfunc = coll.dobj[wsm][key_model]['keys']
        for ii, ff in enumerate(lfunc):
            coll2.add_data(
                ff,
                data=dout[key][ii, ...],
                ref=tuple(ref[1:]),
                **{k0: dout[k0] for k0 in ['dim', 'quant', 'units']},
            )

        # sum
        ksum = f"{key}_sum"
        coll2.add_data(
            ksum,
            data=np.sum(dout[key], axis=0),
            ref=tuple(ref[1:]),
            **{k0: dout[k0] for k0 in ['dim', 'quant', 'units']}
        )

    else:
        ksum = f"{key}_sum"
        coll2.add_data(
            ksum,
            ref=tuple(ref),
            **{k0: dout[k0] for k0 in lk}
        )

    # -------------------------
    # all other vectors if any

    if keyY is not None:
        coll2.add_data(
            keyY,
            **{k0: coll.ddata[keyY][k0] for k0 in lk + ['ref']},
        )

    # --------------
    # add std
    # --------------

    if dout['data_min'] is not None:

        ksmin = f"{key}_sum_min"
        coll2.add_data(
            ksmin,
            data=dout['data_min'],
            ref=tuple(ref),
            **{k0: dout[k0] for k0 in lk[1:]}
        )

        ksmax = f"{key}_sum_max"
        coll2.add_data(
            ksmax,
            data=dout['data_max'],
            ref=tuple(ref),
            **{k0: dout[k0] for k0 in lk[1:]}
        )

    # --------------
    # dkeys
    # --------------

    dkeys = {
        'lamb': key_lamb,
        'sum': ksum,
    }
    if details is True:
        dkeys['details'] = lfunc
    if dout['data_min'] is not None:
        dkeys['sum_min'] = ksmin
        dkeys['sum_max'] = ksmax

    # --------------
    # prepare data
    # --------------

    # get ndim
    ndim = coll2.ddata[dkeys['sum']]['data'].ndim
    if ndim > 2:
        raise NotImplementedError()

    return coll2, dkeys, ndim


# ###############################################################
# ###############################################################
#               plot 1d
# ###############################################################


def _plot_1d(coll2=None, dout=None, dkeys=None, dax=None, details=None):

    # ------------
    # plot
    # -----------

    kax = 'spectrum'
    lax = [vax['handle'] for vax in dax.values() if kax in vax['type']]
    for ax in lax:

        ax.plot(
            coll2.ddata[dkeys['lamb']]['data'],
            coll2.ddata[dkeys['sum']]['data'],
            ls='-',
            marker='None',
            lw=1.,
            color='k',
        )

        # details
        if details is True:
            for ff in dkeys['details']:
                ax.plot(
                    coll2.ddata[dkeys['lamb']]['data'],
                    coll2.ddata[ff]['data'],
                    ls='-',
                    marker='None',
                    lw=1.,
                )

    return dax


# ###############################################################
# ###############################################################
#               plot 2d
# ###############################################################


def _plot_2d(
    coll=None,
    key_model=None,
    coll2=None,
    dout=None,
    dkeys=None,
    keyY=None,
    dax=None,
    details=None,
):

    # --------------
    # plot fixed 2d
    # --------------

    coll2, dgroup = coll2.plot_as_array(
        key=dkeys['sum'],
        keyX=dkeys['lamb'],
        keyY=keyY,
        dax=dax,
        aspect='auto',
        connect=False,
        inplace=True,
    )

    # --------------
    # plot spectrum
    # --------------

    if details is True:

        lamb = coll2.ddata[dkeys['lamb']]['data']
        nmax = dgroup['X']['nmax']
        wsm = coll._which_model
        lfunc = coll.dobj[wsm][key_model]['keys']
        refs = (coll2.ddata[dkeys['sum']]['ref'][0],)
        nan = np.full(lamb.shape, np.nan)

        axtype = 'horizontal'
        lax = [kax for kax, vax in dax.items() if axtype in vax['type']]
        for kax in lax:
            ax = dax[kax]['handle']
            for ii, ff in enumerate(lfunc):

                for jj in range(nmax):

                    ll, = ax.plot(
                        lamb,
                        nan,
                        ls='-',
                        marker='None',
                        lw=1.,
                    )

                    xydata = 'ydata'
                    km = f'{lfunc[ii]}_{jj}'

                    coll2.add_mobile(
                        key=km,
                        handle=ll,
                        refs=(refs,),
                        data=(lfunc[ii],),
                        dtype=[xydata],
                        group_vis='Y',  # 'X' <-> 'Y'
                        axes=kax,
                        ind=jj,
                    )

    return coll2, dgroup


# ###############################################################
# ###############################################################
#               dax
# ###############################################################


def _get_dax(
    ndim=None,
    # figure
    fs=None,
    dmargin=None,
    tit=None,
):

    if ndim == 1:
        return _get_dax_1d(
            fs=fs,
            dmargin=dmargin,
            tit=tit,
        )

    if ndim == 2:
        return _get_dax_2d(
            fs=fs,
            dmargin=dmargin,
            tit=tit,
        )


def _get_dax_1d(
    fs=None,
    dmargin=None,
    tit=None,
):

    # ---------------
    # check inputs
    # ---------------

    if fs is None:
        fs = (10, 6)

    if dmargin is None:
        dmargin = {
            'left': 0.10, 'right': 0.90,
            'bottom': 0.1, 'top': 0.90,
            'wspace': 0.1, 'hspace': 0.1,
        }

    # ---------------
    # prepare figure
    # ---------------

    fig = plt.figure(figsize=fs)
    gs = gridspec.GridSpec(1, 1, **dmargin)

    # ------------
    # add axes
    # ------------

    ax = fig.add_subplot(gs[0, 0])
    # ax.set_xlabel()
    # ax.set_ylabel()

    # ------------
    # populate dax
    # ------------

    dax = {'1d': {'handle': ax, 'type': 'spectrum'}}

    return dax


def _get_dax_2d(
    fs=None,
    dmargin=None,
    tit=None,
):

    # ---------------
    # check inputs
    # ---------------

    if fs is None:
        fs = (18, 10)

    if dmargin is None:
        dmargin = {
            'left': 0.07, 'right': 0.98,
            'bottom': 0.08, 'top': 0.90,
            'wspace': 0.50, 'hspace': 0.10,
        }

    # ---------------
    # prepare figure
    # ---------------

    dax = {}
    fig = plt.figure(figsize=fs)
    gs = gridspec.GridSpec(2, 7, **dmargin)

    # ------------
    # add axes
    # ------------

    ax = fig.add_subplot(gs[:, 0])
    # ax.set_xlabel()
    # ax.set_ylabel()
    dax['vert'] = {'handle': ax, 'type': 'vertical'}

    ax = fig.add_subplot(gs[:, 1:3])
    # ax.set_xlabel()
    # ax.set_ylabel()
    dax['2d'] = {'handle': ax, 'type': 'matrix'}

    ax = fig.add_subplot(gs[0, 3:])
    # ax.set_xlabel()
    # ax.set_ylabel()
    dax['hor'] = {'handle': ax, 'type': 'horizontal'}

    return dax



# ###############################################################
# ###############################################################
#               Finalize figure
# ###############################################################


def _finalize_figure(dax=None, dout=None, tit=None):

    # -------------
    # tit
    # -------------

    titdef = (
        f"Spectral model '{dout['key_model']}'\n"
        f"using data '{dout['key_data']}'"
    )
    tit = ds._generic_check._check_var(
        tit, 'tit',
        types=str,
        default=titdef,
    )

    # -------------
    # tit
    # -------------

    if isinstance(dax, dict):
        fig = list(dax.values())[0]['handle'].figure
    else:
        fig = list(dax.dax.values())[0]['handle'].figure

    if tit is not None:
        fig.suptitle(tit, size=12, fontweight='bold')

    return


import sys
import os
import argparse


# ########################################################
# ########################################################
#               DEFAULTS
# ########################################################


_PKG = 'spectrally'


_HERE = os.path.abspath(os.path.dirname(__file__))
_PATH_PKG = os.path.dirname(os.path.dirname(_HERE))
_PATH_LOCALDIR = os.path.join(os.path.expanduser('~'), f'.{_PKG}')


# ########################################################
# ########################################################
#               get mods
# ########################################################


def get_mods():
    """ Test if git repo """

    # ----------------------------------
    # isgit repo? (load local vs global)

    isgit = False
    if '.git' in os.listdir(_PATH_PKG) and _PKG in _PATH_PKG:
        isgit = True

    if isgit:
        # Make sure we load the corresponding pkg
        sys.path.insert(1, _PATH_PKG)
        import spectrally as sp
        _ = sys.path.pop(1)
    else:
        import spectrally as sp

    # ------------------
    # default parameters

    pfe = os.path.join(_PATH_LOCALDIR, '_scripts_def.py')
    if os.path.isfile(pfe):
        # Make sure we load the user-specific file
        # sys.path method
        # sys.path.insert(1, os.path.join(os.path.expanduser('~'), '.tofu'))
        # import _scripts_def as _defscripts
        # _ = sys.path.pop(1)
        # importlib method
        import importlib.util
        spec = importlib.util.spec_from_file_location("_defscripts", pfe)
        _defscripts = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(_defscripts)
    else:
        try:
            import spectrally.entrypoints._def as _defscripts
        except Exception:
            from . import _def as _defscripts

    return sp, _defscripts


# ########################################################
# ########################################################
#       utility functions
# ########################################################


def _str2bool(v):
    if isinstance(v, bool):
        return v
    elif v.lower() in ['yes', 'true', 'y', 't', '1']:
        return True
    elif v.lower() in ['no', 'false', 'n', 'f', '0']:
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected!')


def _str2boolstr(v):
    if isinstance(v, bool):
        return v
    elif isinstance(v, str):
        if v.lower() in ['yes', 'true', 'y', 't', '1']:
            return True
        elif v.lower() in ['no', 'false', 'n', 'f', '0']:
            return False
        elif v.lower() == 'none':
            return None
        else:
            return v
    else:
        raise argparse.ArgumentTypeError('Boolean, None or str expected!')


def _str2tlim(v):
    c0 = (v.isdigit()
          or ('.' in v
              and len(v.split('.')) == 2
              and all([vv.isdigit() for vv in v.split('.')])))
    if c0 is True:
        v = float(v)
    elif v.lower() == 'none':
        v = None
    return v


# ########################################################
# ########################################################
#       Parser for version
# ########################################################


def parser_version():

    # -------------
    # defaults
    # -------------

    ddef = {
        'path': os.path.join(_PATH_PKG, _PKG),
        'envvar': False,
        'verb': True,
        'warn': True,
        'force': False,
        'name': f'{_PKG.upper()}_VERSION',
    }

    # ------------------
    # Instanciate parser
    # ------------------

    # description msg
    msg = (
        "Get pkg version from bash optionally set an enviroment variable\n\n"
        "If run from a git repo containing {_PKG}, returns git describe\n"
        "Otherwise reads the {_PKG} version stored in {_PKG}/version.py\n"
    )

    # Instanciate parser
    parser = argparse.ArgumentParser(description=msg)

    # ----------------------
    # Define input arguments
    # ----------------------

    # path
    parser.add_argument(
        '-p', '--path',
        type=str,
        help=f'{_PKG} source directory to version.py',
        required=False, default=ddef['path'],
    )

    # verb
    parser.add_argument(
        '-v', '--verb',
        type=_str2bool,
        help='flag indicating whether to print the version',
        required=False,
        default=ddef['verb'],
    )

    # env variable
    parser.add_argument(
        '-ev', '--envvar',
        type=_str2boolstr,
        help='name of the environment variable to set, if any',
        required=False,
        default=ddef['envvar'],
    )

    # warnings
    parser.add_argument(
        '-w', '--warn',
        type=_str2bool,
        help=(
            'flag indicatin whether to print a warning when '
            'desired environment variable (envvar) already exists'
        ),
        required=False,
        default=ddef['warn'],
    )

    # force
    parser.add_argument(
        '-f', '--force',
        type=_str2bool,
        help=(
            'flag indicating whether to force the update of '
            'desired environment variable (envvar) even if it already exists'
        ),
        required=False,
        default=ddef['force'],
    )

    return ddef, parser


# ########################################################
# ########################################################
#       Parser for custom
# ########################################################


def parser_custom():

    # -------------
    # defaults
    # -------------

    ddef = {
        'target': _PATH_LOCALDIR,
        'source': os.path.join(_PATH_PKG, _PKG),
        'files': [
            '_entrypoints_def.py',
        ],
        'directories': [
            'openadas',
            'nist',
            os.path.join('nist', 'ASD'),
        ],
    }

    # ------------------
    # Instanciate parser
    # ------------------

    # description msg
    msg = (
        f" Create a local copy of {_PKG} default parameters\n\n"
        f"A directory '.{_PKG}' is created in your home directory:\n"
        f"\t{_PATH_LOCALDIR}\n"
        "In this directory, modules containing default parameters are copied\n"
        "You can then customize them without impacting other users\n"
    )

    # instanciate
    parser = argparse.ArgumentParser(description=msg)

    # ----------------------
    # Define input arguments
    # ----------------------

    # source
    parser.add_argument(
        '-s', '--source',
        type=str,
        help=f'{_PKG} source directory',
        required=False,
        default=ddef['source'],
    )

    # target
    parser.add_argument(
        '-t', '--target',
        type=str,
        help=(f'directory where .{_PKG}/ should be created'
              + ' (default: {})'.format(ddef['target'])),
        required=False,
        default=ddef['target'],
    )

    # files
    parser.add_argument(
        '-f', '--files',
        type=str,
        help='list of files to be copied',
        required=False,
        nargs='+',
        default=ddef['files'],
        choices=ddef['files'],
    )

    return ddef, parser


# ########################################################
# ########################################################
#       Parser dict
# ########################################################


_DPARSER = {
    'version': parser_version,
    'custom': parser_custom,
}



# -*- coding: utf-8 -*-
"""
Created on Thu Feb 15 22:10:49 2024

@author: dvezinet
"""


from . import test_00_openadas
from . import test_01_nist
from . import test_02_SpectralLines
from . import test_03_SpectralModel
from . import test_04_SpectralFit
from ._hxr_pulses import main as fit_hxr_pulses


# #!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
created on tue feb 20 14:44:51 2024

@author: dvezinet
"""


import numpy as np
import datastock as ds


# local
from . import _class01_fit_func_1d as _1d


#############################################
#############################################
#       defaults
#############################################


_DFUNC = {
    '1d': {
        'sum': _1d._get_func_sum,
        'cost': _1d._get_func_cost,
        'details': _1d._get_func_details,
        'jac': _1d._get_func_jacob,
    },
}


#############################################
#############################################
#       main
#############################################


def main(
    coll=None,
    key=None,
    func=None,
):

    # ----------------
    # check inputs
    # ----------------

    key, key_bs, key_model, func = _check(
        coll=coll,
        key=key,
        func=func,
    )

    # ----------------
    # prepare
    # ----------------

    # n_all
    n_all = len(coll.get_spectral_model_variables(
        key=key_model,
        returnas='all',
        concatenate=True,
    )['all'])

    # param_val
    param_val = coll.get_spectral_model_variables(
        key_model,
        returnas='param_value',
        concatenate=True,
    )['param_value']

    # dind
    dind = coll.get_spectral_model_variables_dind(key_model)

    # ----------------
    # dconstraints
    # ----------------

    # dconstraints
    wsm = coll._which_model
    dconstraints = coll.dobj[wsm][key_model]['dconstraints']

    # coefs
    c0 = dconstraints['c0']
    c1 = dconstraints['c1']
    c2 = dconstraints['c2']

    # check is constraints
    no_constraints = (
        np.allclose(c0, 0)
        and c1.shape == (n_all, n_all)
        and np.allclose(c1, np.eye(n_all))
        and c2.shape == (n_all, n_all)
        and np.allclose(c2, np.eye(n_all))
    )
    if no_constraints:
        c0, c1, c2 = None, None, None

    # ----------------
    # get func
    # ----------------

    if key_bs is None:
        dout = {
            k0: _DFUNC['1d'][k0](
                c0=c0,
                c1=c1,
                c2=c2,
                dind=dind,
                param_val=param_val,
            )
            for k0 in func
        }

    else:
        dout = {
            k0: _DFUNC['2d'][k0](
                coll=coll,
                key=key,
            )
            for k0 in func
        }

    # ----------------
    # return
    # ----------------

    return dout


#############################################
#############################################
#       check
#############################################


def _check(
    coll=None,
    key=None,
    func=None,
):

    # -------------
    # key
    # -------------

    wsm = coll._which_model
    wsf = coll._which_fit
    lok_m = list(coll.dobj.get(wsm, {}).keys())
    lok_1d = [
        k0 for k0, v0 in coll.dobj.get(wsf, {}).items()
        if v0['key_bs'] is None
    ]
    lok_2d = [
        k0 for k0, v0 in coll.dobj.get(wsf, {}).items()
        if v0['key_bs'] is not None
    ]
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok_m + lok_1d + lok_2d,
    )

    # key_bs
    key_bs = None
    if key in lok_2d:
        key_bs = coll.dobj[wsf][key]['key_bs']

    # key_model
    if key in lok_m:
        key_model = key
    else:
        key_model = coll.dobj[wsf][key]['key_model']

    # -------------
    # func
    # -------------

    lok = list(_DFUNC['1d'].keys())
    if func is None:
        func = lok[0]
    if isinstance(func, str):
        func = [func]

    func = ds._generic_check._check_var_iter(
        func, 'func',
        types=list,
        types_iter=str,
        allowed=lok,
    )

    return key, key_bs, key_model, func



# -*- coding: utf-8 -*-
"""
Created on Thu Feb 15 21:53:45 2024

@author: dvezinet
"""
# Built-in
import os


# Standard
import numpy as np
import matplotlib.pyplot as plt


# spectrally-specific
from ._setup_teardown import setup_module0, teardown_module0
from .._class00_SpectralLines import SpectralLines as Collection
from .._saveload import load


_PATH_HERE = os.path.dirname(__file__)
_PATH_INPUT = os.path.join(_PATH_HERE, 'input')
_PATH_OUTPUT = os.path.join(_PATH_HERE, 'output')


#######################################################
#
#     Setup and Teardown
#
#######################################################


def setup_module(module):
    setup_module0(module)


def teardown_module(module):
    teardown_module0(module)


#######################################################
#
#     Instanciate and populate
#
#######################################################


class Test00_Populate():

    # ------------------------
    #   setup and teardown
    # ------------------------

    @classmethod
    def setup_class(cls):
        pass

    def setup_method(self):
        self.coll = Collection()
        self.pfe_json = os.path.join(_PATH_INPUT, 'spectrallines.json')

    # ------------------------
    #   Populating
    # ------------------------

    def test01_add_spectral_lines_from_file(self):
        self.coll.add_spectral_lines_from_file(self.pfe_json)

    def test02_add_spectral_lines_from_openadas(self):
        pass
        # self.coll.add_spectral_lines_from_openadas(
        #     lambmin=3.94e-10,
        #     lambmax=4e-10,
        #     element='Ar',
        #     online=True,
        # )

    def test03_add_spectral_lines_from_nist(self):
        try:
            self.coll.add_spectral_lines_from_nist(
                lambmin=3.94e-10,
                lambmax=4e-10,
                element='Ar',
            )
        except Exception as err:
            lstr = [
                '503 Server Error: Service Unavailable for url:',
                'File could not be downloaded:',
                '=> Maybe check internet connection?',
                # flag that it is running on Github
                (
                    '/runner/.spectrally/nist/',       # MacOS and linux
                    'runneradmin',    # Windows
                ),
            ]
            din = {
                ii: ss in str(err) if isinstance(ss, str)
                else any([s2 in str(err) for s2 in ss])
                for ii, ss in enumerate(lstr)
            }
            if all([vv for vv in din.values()]):
                pass
            else:
                lstr = [f"\t- {k0}: {v0}" for k0, v0 in din.items()]
                msg = "\n\n" + "\n".join(lstr) + "\n"
                raise Exception(msg) from err

    # ----------------
    # removing
    # ----------------

    def test04_remove_spectral_lines(self):
        # populate
        wsl = self.coll._which_lines
        if len(self.coll.dobj.get(wsl, {})) == 0:
            self.coll.add_spectral_lines_from_file(self.pfe_json)

        # remove
        lines = [
            k0 for k0, v0 in self.coll.dobj[wsl].items()
            if v0['ion'] != 'Ar16+'
        ]
        self.coll.remove_spectral_lines(lines)


#######################################################
#
#     Manipulate
#
#######################################################


class Test01_Manipulate():

    # ------------------------
    #   setup and teardown
    # ------------------------

    @classmethod
    def setup_class(cls):
        pass

    def setup_method(self):
        self.coll = Collection()
        self.pfe_json = os.path.join(_PATH_INPUT, 'spectrallines.json')
        self.coll.add_spectral_lines_from_file(self.pfe_json)

        # add spectral data
        t = np.linspace(0, 1, 31)
        lamb = np.linspace(3.94e-10, 4e-10, 100)
        emis1d = np.random.random((lamb.size,))
        emis2d =np.random.random((t.size, lamb.size))
        self.coll.add_data('lamb', lamb, ref='nlamb')
        self.coll.add_data('t', t, ref='nt')
        self.coll.add_data('emis1d', data=emis1d, ref='nlamb')
        self.coll.add_data('emis2d', data=emis2d, ref=('nt', 'nlamb'))

    # ------------------------
    #   Plotting
    # ------------------------

    def test00_plot_spectral_lines(self):
        self.coll.plot_spectral_lines()
        plt.close('all')

    def test01_plot_spectral_data(self):
        key_lines = ['l00', 'l01']
        self.coll.plot_spectral_data(
            'emis1d',
            key_lamb='lamb',
            key_lines=key_lines,
        )
        plt.close('all')

        self.coll.plot_spectral_data(
            'emis2d',
            key_lamb='lamb',
            keyY='t',
            key_lines=key_lines,
        )
        plt.close('all')

    # ------------------------
    #   saving / loading
    # ------------------------

    def test98_save_spectral_lines_to_file(self):
        self.coll.save_spectral_lines_to_file(path=_PATH_OUTPUT)

    def test99_saveload(self):
        pfe = self.coll.save(path=_PATH_OUTPUT, return_pfe=True)
        coll2 = load(pfe, cls=Collection)
        assert self.coll == coll2